import os

from pyfileconf.sectionpath.sectionpath import SectionPath
from tests.input_files.mypackage.cmodule import ExampleClass
from tests.test_pipeline_manager.test_create import PipelineManagerLoadTestBase

from pyfileconf import Selector
from tests.input_files.amodule import a_function
from tests.test_pipeline_manager.base import PipelineManagerTestBase, CLASS_CONFIG_DICT_LIST


class PipelineManagerCreateEntryTestBase(PipelineManagerLoadTestBase):
    pass


class TestPipelineManagerCreateEntry(PipelineManagerCreateEntryTestBase):

    def test_create_entry_for_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager.create('thing', a_function)
        sel = Selector()
        iv = sel.test_pipeline_manager.thing.a_function
        module_folder = os.path.join(self.defaults_path, 'thing')
        function_path = os.path.join(module_folder, 'a_function.py')
        with open(function_path, 'r') as f:
            contents = f.read()
            self.assert_a_function_config_file_contents(contents)

    def test_create_entries_for_function_get_section(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager.create('thing.stuff', a_function)
        pipeline_manager.create('thing.woo', a_function)
        sel = Selector()
        base_folder = os.path.join(self.defaults_path, 'thing')
        module_folders = [os.path.join(base_folder, attr) for attr in ['stuff', 'woo']]
        for module_folder in module_folders:
            function_path = os.path.join(module_folder, 'a_function.py')
            with open(function_path, 'r') as f:
                contents = f.read()
                self.assert_a_function_config_file_contents(contents)
        iv = sel.test_pipeline_manager.thing
        result = pipeline_manager.get(iv)
        assert result['stuff'][0].func == a_function
        assert len(result['stuff']) == 1
        assert result['woo'][0].func == a_function
        assert len(result['woo']) == 1
        assert len(result) == 2

    def test_create_entry_for_previously_unimported_function(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager.create('thing', a_function)
        sel = Selector()
        iv = sel.test_pipeline_manager.thing.a_function
        module_folder = os.path.join(self.defaults_path, 'thing')
        function_path = os.path.join(module_folder, 'a_function.py')
        with open(function_path, 'r') as f:
            contents = f.read()
            self.assert_a_function_config_file_contents(contents)
        module_folder = os.path.join(self.defaults_path, 'stuff')
        class_path = os.path.join(module_folder, 'ExampleClass.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_config_file_contents(contents)

    def test_update_then_create_entry_for_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        orig_iv = sel.test_pipeline_manager.stuff.a_function
        expected_b_result = ['a', 'b']
        section_path = SectionPath.from_section_str_list(SectionPath(orig_iv.section_path_str)[1:])
        pipeline_manager.update(
            b=expected_b_result,
            section_path_str=section_path.path_str
        )
        pipeline_manager.create('thing', a_function)
        sel = Selector()
        iv = sel.test_pipeline_manager.thing.a_function
        module_folder = os.path.join(self.defaults_path, 'thing')
        function_path = os.path.join(module_folder, 'a_function.py')
        with open(function_path, 'r') as f:
            contents = f.read()
            self.assert_a_function_config_file_contents(contents)
        result = pipeline_manager.run(orig_iv)
        assert result == (None, expected_b_result)

    def test_create_then_update_entry_for_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager.create('thing', a_function)
        sel = Selector()
        iv = sel.test_pipeline_manager.thing.a_function
        expected_b_result = ['a', 'b']
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            b=expected_b_result,
            section_path_str=section_path.path_str
        )
        module_folder = os.path.join(self.defaults_path, 'thing')
        function_path = os.path.join(module_folder, 'a_function.py')
        with open(function_path, 'r') as f:
            contents = f.read()
            self.assert_a_function_config_file_contents(contents)
        result = pipeline_manager.run(iv)
        assert result == (None, expected_b_result)

    def test_create_entry_for_class(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager.create('thing', ExampleClass)
        sel = Selector()
        iv = sel.test_pipeline_manager.thing.ExampleClass
        module_folder = os.path.join(self.defaults_path, 'thing')
        class_path = os.path.join(module_folder, 'ExampleClass.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_config_file_contents(contents)

    def test_create_entries_for_class_get_section(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager.create('thing.one', ExampleClass)
        pipeline_manager.create('thing.two', ExampleClass)
        sel = Selector()
        base_folder = os.path.join(self.defaults_path, 'thing')
        module_folders = [os.path.join(base_folder, attr) for attr in ['one', 'two']]
        for module_folder in module_folders:
            class_path = os.path.join(module_folder, 'ExampleClass.py')
            with open(class_path, 'r') as f:
                contents = f.read()
                self.assert_example_class_config_file_contents(contents)
        iv = sel.test_pipeline_manager.thing
        result = pipeline_manager.get(iv)
        assert result['one'][0].func == ExampleClass
        assert len(result['one']) == 1
        assert result['two'][0].func == ExampleClass
        assert len(result['two']) == 1
        assert len(result) == 2


    def test_create_entry_for_previously_unimported_class(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager.create('thing', ExampleClass)
        sel = Selector()
        iv = sel.test_pipeline_manager.thing.ExampleClass
        module_folder = os.path.join(self.defaults_path, 'stuff')
        function_path = os.path.join(module_folder, 'a_function.py')
        with open(function_path, 'r') as f:
            contents = f.read()
            self.assert_a_function_config_file_contents(contents)
        module_folder = os.path.join(self.defaults_path, 'thing')
        class_path = os.path.join(module_folder, 'ExampleClass.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_config_file_contents(contents)

    def test_create_entry_for_specific_class_dict(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager.create('example_class.thing.data')
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.thing.data
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'thing')
        class_path = os.path.join(module_folder, 'data.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(contents)

    def test_create_entries_for_specific_class_dict_get_section(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager.create('example_class.thing.data')
        pipeline_manager.create('example_class.thing.data2')
        sel = Selector()
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'thing')
        for attr in ['data', 'data2']:
            class_path = os.path.join(module_folder, f'{attr}.py')
            with open(class_path, 'r') as f:
                contents = f.read()
                self.assert_example_class_dict_config_file_contents(
                    contents, name_value= f"name: Optional[str] = '{attr}'"
                )
        iv = sel.test_pipeline_manager.example_class.thing
        result = pipeline_manager.get(iv)
        assert result[0] == ExampleClass(a=None, name='data')
        assert result[1] == ExampleClass(a=None, name='data2')
        assert len(result) == 2


    def test_update_then_create_entry_for_specific_class_dict(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        orig_iv = sel.test_pipeline_manager.example_class.stuff.data
        expected_a_result = (1, 2)
        section_path = SectionPath.from_section_str_list(SectionPath(orig_iv.section_path_str)[1:])
        pipeline_manager.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        pipeline_manager.create('example_class.thing.data')
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.thing.data
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'thing')
        class_path = os.path.join(module_folder, 'data.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(contents)
        ec = sel.test_pipeline_manager.example_class.stuff.data
        expect_ec = ExampleClass(name='data', a=expected_a_result)
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

    def test_create_then_update_entry_for_specific_class_dict(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager.create('example_class.thing.data')
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.thing.data
        expected_a_result = (1, 2)
        section_path = SectionPath.from_section_str_list(SectionPath(iv.section_path_str)[1:])
        pipeline_manager.update(
            a=expected_a_result,
            section_path_str=section_path.path_str
        )
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'thing')
        class_path = os.path.join(module_folder, 'data.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(contents)

        ec = sel.test_pipeline_manager.example_class.thing.data
        result = pipeline_manager.run(ec)
        assert result == 'woo'

        expect_ec = ExampleClass(name='data', a=expected_a_result)
        got_ec = ec.item
        assert ec.name == expect_ec.name == got_ec.name
        assert ec.a == expect_ec.a == got_ec.a



    def test_create_deeply_nested_entry_for_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager.create('thing.stuff.whoa', a_function)
        sel = Selector()
        iv = sel.test_pipeline_manager.thing.stuff.whoa.a_function
        module_folder = os.path.join(self.defaults_path, 'thing')
        module_folder = os.path.join(module_folder, 'stuff')
        module_folder = os.path.join(module_folder, 'whoa')
        function_path = os.path.join(module_folder, 'a_function.py')
        with open(function_path, 'r') as f:
            contents = f.read()
            self.assert_a_function_config_file_contents(contents)

    def test_create_deeply_nested_entry_for_specific_class_dict(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager.create('example_class.thing.stuff.whoa.data')
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.thing.stuff.whoa.data
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'thing')
        module_folder = os.path.join(module_folder, 'stuff')
        module_folder = os.path.join(module_folder, 'whoa')
        class_path = os.path.join(module_folder, 'data.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(contents)

    def test_create_entry_for_specific_class_dict_in_same_section_as_existing_entry(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager.create('example_class.stuff.data2')
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data2
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'stuff')
        class_path = os.path.join(module_folder, 'data2.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(
                contents,
                name_value="name: Optional[str] = 'data2'"
            )



from typing import Union, Tuple, Dict, Optional
import ast

from pyfileconf.exceptions.assignments import CouldNotExtractAssignmentFromAstException
from pyfileconf.io.file.write.asttosource import ast_node_to_source
from pyfileconf.io.file.load.parsers.assign import extract_assignment_from_ast
from mixins.repr import ReprMixin
from pyfileconf.mixins.orderpref import OrderPreferenceMixin

AnyAstAssign = Union[ast.Assign, ast.AnnAssign]

class AssignmentStatement(ReprMixin, OrderPreferenceMixin):
    repr_cols = ['varname', 'value', 'annotation']

    def __init__(self, target: ast.Name, value: ast.AST, annotation: ast.Name=None,
                 preferred_position: str = None):
        self.target = target
        self.varname = target.id if isinstance(target, ast.Name) else ast_node_to_source(target)  # handle subscripts, etc.
        self.value = value
        self.annotation = annotation
        self.preferred_position = preferred_position  # sets self.prefer_beginning as bool

        # Will get set to True if creating using classmethod self.from_str.
        # When set to True, will simply use the original str to output back to str
        self.created_from_str = False
        self.orig_str: Optional[str] = None

    def __str__(self):
        # Created from str, just return that str back (keeps whitespace, comments, etc.)
        if self.created_from_str:
            return self.orig_str

        # Not created from str, use ast to create str
        ast_assign = self.to_ast()
        return ast_node_to_source(ast_assign)

    def __eq__(self, other):
        if not isinstance(other, AssignmentStatement):
            # Auto convert ast assigns for comparison
            if isinstance(other, (ast.Assign, ast.AnnAssign)):
                other = AssignmentStatement.from_ast_assign(other)
            else:
                return False  # If not an assign or ast assign, not equal

        # Other is an assignment statement.
        if self.varname != other.varname:
            return False  # non-matching assignment variable, must not be equal

        # Other is an assignment statement to the same variable
        # Check if source generated from ast values are the same
        if ast_node_to_source(self.value) != ast_node_to_source(other.value):
            return False

        # Ignore differences on annotations. Could check that here.

        # Passed all checks, must be the same
        return True

    @classmethod
    def from_ast_assign(cls, ast_assign: AnyAstAssign, preferred_position: str = None):

        if isinstance(ast_assign, ast.Assign):
            # Base assignment, no annotation. May be multiple targets
            target = ast_assign.targets[0]

            # TODO [#1]: multiple assignments (multiple targets)
            if len(ast_assign.targets) > 1:
                raise NotImplementedError(f'cannot yet handle multiple assignment to: {ast_assign.targets}')

            annotation = None
        elif isinstance(ast_assign, ast.AnnAssign):
            # Annotation assignment. Must be single target.
            target = ast_assign.target
            annotation = ast_assign.annotation
        else:
            raise ValueError(f'expected ast.Assign or ast.AnnAssign. got {ast_assign} of type {type(ast_assign)}')

        # TODO [#22]: remove type ignores in AssignmentStatement.from_ast_assign once ast has proper type definitions
        return cls(
            target,  # type: ignore
            ast_assign.value,  # type: ignore
            annotation,  # type: ignore
            preferred_position=preferred_position
        )

    def to_default_dict_and_annotation_dict(self) -> Tuple[Dict[str, ast.AST], Dict[str, ast.Name]]:
        # Set default dict
        try:
            default_dict = {self.target.id: self.value}
        except AttributeError as e:
            if "no attribute 'id'" in str(e):
                try:
                    self.target.attr  # type: ignore
                    # Success, means that this was an attribute assignment like a.b = 6, do not need to put
                    # into default dict
                    default_dict = {}
                except AttributeError:
                    raise NotImplementedError(
                        f'do not understand how to parse {self.target} (dict: {self.target.__dict__})'
                    )
            else:
                raise e

        # Set annotation dict
        annotation_dict = {}
        if self.annotation is not None:
            annotation_dict = {self.target.id: self.annotation}

        return default_dict, annotation_dict

    def to_ast(self) -> AnyAstAssign:
        if self.annotation is None:
            return ast.Assign(
                targets=[self.target],
                value=self.value
            )
        else:
            return ast.AnnAssign(
                target=self.target,
                value=self.value,
                annotation=self.annotation,
                simple=1
            )

    @classmethod
    def from_varname_and_ast_value(cls, varname: str, value: ast.AST, annotation: ast.Name=None,
                                   preferred_position: str = None):
        assign_name = ast.Name(id=varname, ctx=ast.Store())

        return cls(assign_name, value=value, annotation=annotation, preferred_position=preferred_position)

    @classmethod
    def from_str(cls, assign_str: str, preferred_position: str = None):
        ast_module = ast.parse(assign_str)
        cls_obj = extract_assignment_from_ast(ast_module)
        if cls_obj is None:
            raise CouldNotExtractAssignmentFromAstException(f'Original str which created ast: {assign_str}')
        cls_obj.preferred_position = preferred_position
        cls_obj.created_from_str = True
        cls_obj.orig_str = assign_str

        return cls_obj


import ast
from typing import List, Optional


class NameExtractor(ast.NodeVisitor):

    def __init__(self):
        self.names = []

    def visit_Name(self, node):
        self.names.append(node.id)


def extract_external_name_from_assign_value(ast_node: ast.AST) -> Optional[str]:
    """
    Identifies the root part of an assignment value, so long as it is not builtin.

    Examples
    from a import b

    b -> 'b'
    b.attr -> 'b'

    import c

    c.test -> 'c'
    c.thing.stuff -> 'c'

    'abc' -> None
    123 -> None

    Args:
        ast_node: should be ast form of the value of an assignment

    Returns:

    """
    ne = NameExtractor()
    ne.visit(ast_node)

    if len(ne.names) == 0:
        return None

    if len(ne.names) > 1:
        raise ValueError(f'expected single name, got {ne.names}')

    return ne.names[0]


def extract_unique_external_names_from_assign_value(ast_node: ast.AST) -> List[str]:
    """
    Same as extract_external_name_from_assign_value but returns a list always, can return multiple names,
    and drops duplicates in the returned names
    """
    ne = NameExtractor()
    ne.visit(ast_node)

    return list(set(ne.names))



import os
import shutil
from typing import Callable, Type, Optional, List, Dict, Union


def delete_project(path: str, logs_path: str,
                   specific_class_config_dicts: Optional[List[Dict[str, Union[str, Type, List[str]]]]] = None,):
    all_paths = [
        os.path.join(path, 'defaults'),
        os.path.join(path, 'custom_defaults'),
        os.path.join(path, 'pipeline_dict.py'),
        logs_path,
    ]
    for specific_class_config in specific_class_config_dicts:
        name = specific_class_config['name']
        all_paths.append(os.path.join(path, f'{name}_dict.py'))
    for path in all_paths:
        if os.path.isdir(path):
            shutil.rmtree(path)
        elif os.path.isfile(path):
            os.remove(path)
        else:
            # Must not exist, maybe add handling for this later
            pass


def pipeline_dict_str_with_obj(func: Callable, func_key: str, func_module: str) -> str:
    return f'from {func_module} import {func.__name__}\n\npipeline_dict = {{\n\t"{func_key}": [{func.__name__}],\n}}\n'


def nested_pipeline_dict_str_with_obj(func: Callable, section_key: str, func_key: str, func_module: str) -> str:
    return f'from {func_module} import {func.__name__}\n\npipeline_dict = {{\n\t"{section_key}": {{\n\t\t"{func_key}": [{func.__name__}],\n\t}},\n}}\n'


def class_dict_str(name: str, key: str, value: str) -> str:
    return f'\n{name} = {{\n\t"{key}": ["{value}"],\n}}\n'


def nested_class_dict_str(name: str, section_key: str, key: str, value: str) -> str:
    return f'\n{name} = {{\n\t"{section_key}": {{\n\t\t"{key}": ["{value}"],\n\t}},\n}}\n'




"""
py-file-conf is a Python framework for flow-based programming and managing configuration. To use it, you register
your main functions and classes, and config files are created automatically for them. It provides a way to run these functions
individually, in a list, in sections, or a combination thereof. Configuration can be 
dynamically updated, enabling powerful scripting.
"""
from pyfileconf.plugin.impl import hookimpl
from pyfileconf.main import PipelineManager
from pyfileconf.selector.models.selector import Selector
from pyfileconf.iterate import IterativeRunner
from pyfileconf.plugin.manager_utils import reset_plugins, remove_default_plugins
from pyfileconf.context import context



import os
from typing import Union, Any, Optional

from mixins.repr import ReprMixin

from pyfileconf.config.models.file import ActiveFunctionConfigFile
from pyfileconf.exceptions.config import ConfigManagerNotLoadedException
from pyfileconf.logic.get import _get_from_nested_obj_by_section_path
from pyfileconf.logic.set import _set_in_nested_obj_by_section_path
from pyfileconf.config.models.interfaces import ConfigSectionOrConfig
from pyfileconf.config.models.section import ConfigSection, ActiveFunctionConfig
from pyfileconf.pipelines.models.file import FunctionConfigFile
from pyfileconf.sectionpath.sectionpath import SectionPath

class ConfigManager(ReprMixin):
    repr_cols = ['basepath', 'section']

    def __init__(self, basepath: str, main_section: ConfigSection=None):
        self.section = main_section
        self.basepath = basepath
        self.local_config = ActiveFunctionConfig()

    def __getattr__(self, item):
        return getattr(self.section, item)

    def __dir__(self):
        exposed_methods = [
            'load',
            'get',
            'update'
        ]
        exposed_attrs = [
            'basepath',
        ]
        return exposed_methods + exposed_attrs + list(self.section.config_map.keys())

    def load(self):
        self.section = ConfigSection.from_files(self.basepath)

    def update(self, d: dict=None, section_path_str: str=None, **kwargs) -> None:
        config_obj = self._get_project_config_or_local_config_by_section_path(section_path_str)
        if config_obj is None:
            raise ConfigManagerNotLoadedException('no config to update')
        config_obj.update(d, **kwargs)

    def refresh(self, section_path_str: str):
        config_obj = self._get_project_config_or_local_config_by_section_path(section_path_str)
        if config_obj is None:
            raise ConfigManagerNotLoadedException('no config to refresh')
        config_obj.refresh()

    def reset(self, section_path_str: str=None, allow_create: bool = False) -> None:
        """
        Resets a function or section config to default. If no section_path_str is passed, resets local config.

        To reset all configs, use .load() instead.

        Args:
            section_path_str:

        Returns:

        """
        default = self._get_default_func_or_section_config(section_path_str, create=allow_create)
        self.set(section_path_str, default, allow_create=allow_create)

    def pop(self, key: str, section_path_str: str=None) -> Any:
        config_obj = self._get_project_config_or_local_config_by_section_path(section_path_str)
        if config_obj is None:
            raise ConfigManagerNotLoadedException('no config to pop')
        return config_obj.pop(key)

    def get(self, section_path_str: str) -> Optional[ActiveFunctionConfig]:
        """
        Handles config inheritance to get the active config for a section or function

        Args:
            section_path_str:

        Returns:

        """
        config = self._get_func_or_section_configs(section_path_str)

        if self.section is None:
            raise ConfigManagerNotLoadedException('call .load() on ConfigManager before .get()')

        # First override for function defaults is global project config
        section_configs = [self.section.config]

        # Get configs, in order of highest level to lowest level. Will go from project to highest section,
        # down to lowest section.
        section_path = SectionPath(section_path_str)
        full_section = ''
        for section in section_path[:-1]: # skip the last section or function for special handling at end
            full_section += section # rebuilding full section path str
            section_configs.append(
                self._get_func_or_section_configs(full_section)
            )
            full_section += '.'

        # Last item of section_path may be another section, or the function/Pipeline itself. If it's a section,
        # must add config for override, but if is function, it is already the base config so should not update.
        full_section += section_path[-1]
        if not self._is_function_or_pipeline_path(full_section):
            # if is a section, not function/pipeline
            section_configs.append(self._get_func_or_section_configs(full_section))

        if config:
            # Override configs. Default config is base config, then gets updated by project, then high
            # level sections to low level sections
            [config.update(section_config) for section_config in section_configs]

            # Last, override with local config
            config.update(self.local_config)

        return config

    def set(self, section_path_str: str=None, value=None, allow_create: bool = True):
        """
        In contrast to update, completely replaces the config object.

        Args:
            section_path_str:
            value:

        Returns:

        """

        if value == None: # empty config
            value = ActiveFunctionConfig()

        if section_path_str is None:
            # updating local config
            self.local_config = value
            return

        self._set_func_or_section_config(section_path_str, value=value, allow_create=allow_create)


    def _get_func_or_section_configs(self, section_path_str: str) -> Optional[ActiveFunctionConfig]:
        """
        This get method is used to get only the config for the section path, without handling
        multiple levels of config and overriding. To get the active config for a function,
        use regular get method.

        Args:
            section_path_str:

        Returns:

        """
        if self.section is None:
            raise ConfigManagerNotLoadedException('call .load() on ConfigManager before .get()')

        if section_path_str is None:
            section_path_str = self.section.name

        section_path = SectionPath(section_path_str)

        # Goes into nested sections, until it pulls the final config or section
        config_or_section: ConfigSectionOrConfig = _get_from_nested_obj_by_section_path(self, section_path)
        conf = _get_config_from_config_or_section(config_or_section)

        # Now update stored config as loading may have happened during _get_config_from_config_or_section
        # Want to keep the active config once it is loaded
        # But if it is a section, then don't want to overwrite with config
        if not isinstance(config_or_section, ConfigSection):
            _set_in_nested_obj_by_section_path(self, section_path, conf)

        return conf

    def _set_func_or_section_config(self, section_path_str: str, value=None, allow_create: bool = True) -> None:
        if self.section is None:
            raise ConfigManagerNotLoadedException('call .load() on ConfigManager before .set()')

        if section_path_str is None:
            section_path_str = self.section.name

        section_path = SectionPath(section_path_str)

        if allow_create:
            self._set_func_or_config_with_create(section_path, value)
        else:
            self._set_func_or_config_no_create(section_path, value)

    def _set_func_or_config_with_create(self, section_path: SectionPath, value: Any):
        obj = self
        section_basepath = self.basepath
        for i, section in enumerate(section_path[:-1]):
            section_basepath = os.path.join(section_basepath, section)
            try:
                obj = getattr(obj, section)
            except KeyError as e:
                new_section = ConfigSection([], name=section)
                obj.append(new_section)
                obj = getattr(obj, section)

        # Now have collection object which should hold this final object
        obj.append(value)

    def _set_func_or_config_no_create(self, section_path: SectionPath, value: Any):
        _set_in_nested_obj_by_section_path(self, section_path, value)

    def _get_default_func_or_section_config(self, section_path_str: str=None,
                                            create: bool = False) -> Union[ActiveFunctionConfig, ConfigSection]:

        if section_path_str is None:
            # local config. Default is blank config
            return ActiveFunctionConfig()
        else:
            # otherwise, load from file for default
            section_path = SectionPath(section_path_str)
            filepath = section_path.to_filepath(self.basepath)

            try:
                config_obj = _get_from_nested_obj_by_section_path(self, section_path)
            except KeyError as e:
                # config object not already created
                if not create:
                    raise e
                config_obj = ActiveFunctionConfig()
            if isinstance(config_obj, ConfigSection):
                return ConfigSection.from_files(filepath)
            if isinstance(config_obj, (ActiveFunctionConfig, ActiveFunctionConfigFile)):
                return ActiveFunctionConfig.from_file(filepath + '.py')
            else:
                raise ValueError(f'expected section path to return ConfigSection or FunctionConfig, '
                                 f'got {config_obj} of type {type(config_obj)}')

    def _is_function_or_pipeline_path(self, section_path_str: str) -> bool:
        section_path = SectionPath(section_path_str)
        # Goes into nested sections, until it pulls the final config or section
        config_or_section: ConfigSectionOrConfig = _get_from_nested_obj_by_section_path(self, section_path)
        if isinstance(config_or_section, ConfigSection):
            # must be section, not individual pipeline or function
            return False
        elif isinstance(config_or_section, (ActiveFunctionConfig, ActiveFunctionConfigFile)):
            # must be individual function as Config is returned
            return True
        else:
            raise ValueError(f'expected Config or ConfigSection, got {config_or_section} of type {config_or_section}')

    def _get_project_config_or_local_config_by_section_path(self, section_path_str: Optional[str]
                                                            ) -> Optional[ActiveFunctionConfig]:
        if section_path_str is not None:
            config_obj = self._get_func_or_section_configs(section_path_str)
        else:
            # If no section passed, update local config
            config_obj = self.local_config

        return config_obj


def _get_config_from_config_or_section(config_or_section: ConfigSectionOrConfig) -> Optional[ActiveFunctionConfig]:
    # Pull Config file from ConfigSection
    if isinstance(config_or_section, ActiveFunctionConfig):
        # config already loaded
        return config_or_section
    if isinstance(config_or_section, ConfigSection):
        # config already loaded
        return config_or_section.config
    if isinstance(config_or_section, ActiveFunctionConfigFile):
        # Load config from file
        return config_or_section.load(ActiveFunctionConfig)

    raise ValueError(f'expected Config or ConfigSection, got {config_or_section} of type {config_or_section}')



from typing import List, Union
import os
import warnings

from mixins.repr import ReprMixin
from pyfileconf.basemodels.container import Container
from pyfileconf.config.models.config import ActiveFunctionConfig
from pyfileconf.config.models.file import ActiveFunctionConfigFile
from pyfileconf.exceptions.config import ConfigManagerNotLoadedException
from pyfileconf.pipelines.models.file import FunctionConfigFile
from pyfileconf.sectionpath.sectionpath import _strip_py


class ConfigSection(Container, ReprMixin):
    repr_cols = ['name', 'config', 'items']

    def __init__(self, configs: List[Union[ActiveFunctionConfigFile, 'ConfigSection']],
                 section_config: ActiveFunctionConfig=None, name: str=None):
        self.config = section_config
        self.items = configs
        self.name = name

    def __getattr__(self, item):

        # Handle if trying to get this section's config
        if self.config is not None and item == self.config.name:
            return self.config

        return self.config_map[item]

    def __dir__(self):
        return self.config_map.keys()

    @property  # type: ignore
    def items(self):
        return self._items

    @items.setter
    def items(self, items):
        self._items = items
        self._set_config_map() # recreate config map on setting items

    def append(self, item):
        self._items.append(item)
        self._set_config_map()

    def extend(self, items):
        self._items.extend(items)
        self._set_config_map()

    @property
    def config_map(self):
        return self._config_map

    def _set_config_map(self):
        config_map = {}
        for config in self:
            if config.name is None:
                warnings.warn(f"Couldn't determine name of config {config}. Can't add to mapping.")
                continue
            config_map[config.name] = config
        self._config_map = config_map

    def update(self, d: dict=None, **kwargs):
        if self.config is None:
            raise ConfigManagerNotLoadedException('no config in ConfigSection')
        if d is not None:
            self.config.update(d)
        self.config.update(kwargs)

    @classmethod
    def from_files(cls, basepath: str):

        # Get section name by name of folder
        name = os.path.basename(os.path.abspath(basepath))

        # Config folder doesn't exist, return empty config
        if not os.path.exists(basepath):
            return cls([], name=name)

        config_file_list = [file for file in next(os.walk(basepath))[2] if file.endswith('.py')]

        # Ignore folders starting with . or _
        config_section_list = [
            folder for folder in next(os.walk(basepath))[1] if not any(
                [folder.startswith(item) for item in ('_','.')]
            )
        ]

        # Special handling for section config
        try:
            config_file_list.remove('section.py')
            section_config = ActiveFunctionConfig.from_file(os.path.join(basepath, 'section.py'), name=name)
        except ValueError:
            # Didn't find section config
            section_config = None

        configs = [
            ActiveFunctionConfigFile(
                os.path.join(basepath, file),
                name=_strip_py(file)
            ) for file in config_file_list
        ]
        # Recursively calling section creation to create individual config files
        config_sections = [ConfigSection.from_files(os.path.join(basepath, folder)) for folder in config_section_list]

        configs += config_sections

        return cls(configs, section_config=section_config, name=name)




from pyfileconf.assignments.models.container import AssignmentStatementContainer
from pyfileconf.assignments.models.statement import AssignmentStatement
from pyfileconf.basemodels.config import ConfigBase, ImportsAndAssigns
from pyfileconf.pipelines.models.dictfile import PipelineDictFile


class PipelineDictConfig(ConfigBase):

    config_file_class = PipelineDictFile

    def as_imports_and_assignments(self) -> ImportsAndAssigns:
        pipeline_dict_assign = AssignmentStatement.from_str(self['pipeline_dict'])
        assigns = AssignmentStatementContainer([pipeline_dict_assign])
        return self.imports, assigns





from mixins.repr import ReprMixin

class Comment(ReprMixin):
    repr_cols = ['comment']

    def __init__(self, comment_str: str):
        self.comment = comment_str

    def __str__(self):
        return f'#{self.comment}'



import ast
from typing import Tuple

from pyfileconf.io.func.load.args import FunctionArgsExtractor

class FunctionConfigExtractor(FunctionArgsExtractor):

    def extract_config(self, name: str=None):
        from pyfileconf.pipelines.models.config import FunctionConfig
        args, func_arg_imports = super().extract_args()

        # Parse into config
        defaults_dict, annotation_dict = function_args_as_arg_and_annotation_dict(args)

        return FunctionConfig(
            defaults_dict,
            annotations=annotation_dict,
            name=name,
            imports=func_arg_imports
        )



def function_args_as_arg_and_annotation_dict(args: ast.arguments) -> Tuple[dict, dict]:
    arg_dict = {}
    annotation_dict = {}

    # Handle args
    num_no_default_value = _get_length_of_arg_section(args.args) - _get_length_of_arg_section(args.defaults)
    all_defaults = [ast.NameConstant(value=None)] * num_no_default_value + _get_list_of_arg_section(args.defaults)
    assert len(all_defaults) == _get_length_of_arg_section(args.args)

    for i, arg in enumerate(args.args):
        default = all_defaults[i]
        arg_dict[arg.arg] = default
        if arg.annotation is not None:
            annotation_dict[arg.arg] = _strip_ast_quotes_if_necessary(arg.annotation)

    # Handle kwargs
    for i, kwarg in enumerate(args.kwonlyargs):
        default = args.kw_defaults[i]
        arg_dict[kwarg.arg] = default
        if kwarg.annotation is not None:
            annotation_dict[kwarg.arg] = _strip_ast_quotes_if_necessary(kwarg.annotation)

    return arg_dict, annotation_dict


def _get_length_of_arg_section(section):
    if section is None:
        return 0

    return len(section)


def _get_list_of_arg_section(section):
    if section is None:
        return []

    return list(section)


def _strip_ast_quotes_if_necessary(node: ast.AST) -> ast.AST:
    if isinstance(node, ast.Str):
        # Got a quoted annotation such as 'DataPipeline'. Strip the quotes
        str_value = node.s
        return ast.Name(str_value)
    return node



import ast

ast_none = ast.NameConstant(value=None)
ast_dict_constructor = ast.Call(
    func=ast.Name(id='dict'),
    args=[],
    keywords=[]
)

def ast_str(string: str) -> ast.Str:
    return ast.Str(s=string)

def ast_dict_constructor_with_kwargs_from_dict(d: dict) -> ast.Call:
    return ast.Call(
        func=ast.Name(id='dict'),
        args=[],
        keywords=[
            ast.keyword(
                arg=key,
                value=value
            ) for key, value in d.items()
        ]
    )


import ast
import sys
import os
import traceback
from collections import defaultdict
from functools import partial

import bdb
import warnings
from copy import deepcopy
from typing import TYPE_CHECKING, Union, List, Callable, Tuple, Optional, Any, Sequence, Type, cast, Dict, Set

from pyfileconf.basemodels.registrar import Registrar
from pyfileconf.data.logic.convert import convert_to_empty_obj_if_necessary
from pyfileconf.data.models.collection import SpecificClassCollection
from pyfileconf.data.models.dictconfig import SpecificClassDictConfig
from pyfileconf.debug import pdb_post_mortem_or_passed_debug_fn
from pyfileconf.dictfiles.modify import add_item_into_nested_dict_at_section_path, \
    create_dict_assignment_str_from_nested_dict_with_ast_names, pretty_format_str
from pyfileconf.imports.logic.load.name import get_module_and_name_imported_from
from pyfileconf.imports.models.statements.container import ImportStatementContainer
from pyfileconf.imports.models.statements.obj import ObjectImportStatement
from pyfileconf.iterate import IterativeRunner
from pyfileconf.logic.combine import \
    combine_items_into_list_whether_they_are_lists_or_not_then_extract_from_list_if_only_one_item
from pyfileconf.pipelines.models.collection import PipelineCollection
from pyfileconf.pipelines.models.dictconfig import PipelineDictConfig
from pyfileconf.plugin import manager as plugin_manager
from pyfileconf.views.object import ObjectView

if TYPE_CHECKING:
    from pyfileconf.runner.models.interfaces import RunnerArgs, StrOrView, IterativeResults

from pyfileconf.config.models.manager import ConfigManager
from pyfileconf.pipelines.models.registrar import PipelineRegistrar
from pyfileconf.pipelines.models.dictfile import PipelineDictFile
from pyfileconf.data.models.registrar import SpecificRegistrar
from pyfileconf.data.models.dictfile import SpecificClassDictFile
from pyfileconf.runner.models.runner import Runner, ResultOrResults
from pyfileconf.imports.models.tracker import ImportTracker
from pyfileconf.sectionpath.sectionpath import SectionPath
from pyfileconf.exceptions.pipelinemanager import PipelineManagerNotLoadedException, \
    NoPipelineManagerForFilepathException, NoPipelineManagerForSectionPathException
from pyfileconf.logger import stdout_also_logged
from pyfileconf.interfaces import SpecificClassConfigDict
from pyfileconf.context import context


class PipelineManager:
    """
    Main class for managing flow-based programming and configuration.
    """

    def __init__(self, folder: str,
                 name: str= 'project',
                 specific_class_config_dicts: Optional[List[SpecificClassConfigDict]] = None,
                 auto_pdb: Union[bool, Callable] = False, force_continue: bool = False,
                 log_folder: Optional[str] = None,
                 default_config_folder_name: str = 'defaults'):

        if specific_class_config_dicts is None:
            specific_class_config_dicts = []

        self.folder = folder
        self.pipeline_dict_path = os.path.join(folder, 'pipeline_dict.py')
        self.specific_class_config_dicts = specific_class_config_dicts
        self.specific_class_names = [sc_dict['name'] for sc_dict in specific_class_config_dicts]
        self.default_config_path = os.path.join(self.folder, default_config_folder_name)
        self.name = name
        self.auto_pdb = auto_pdb
        self.force_continue = force_continue
        self.log_folder = log_folder

        self._registrars: Optional[Sequence[Registrar]] = None
        self._general_registrar: Optional[PipelineRegistrar] = None

        self._validate_options()
        context.active_managers[self.name] = self
        self._create_project_if_needed()


    def __getattr__(self, item):

        if item == 'runner':
            # must not be defined yet
            raise PipelineManagerNotLoadedException('call PipelineManager.load() before accessing '
                                                    'functions and classes')

        # Must be getting function
        return getattr(self.runner, item)

    def __dir__(self):
        exposed_methods = [
            'run',
            'run_iter',
            'get',
            'update',
            'load',
            'reload',
            'refresh',
            'reset',
        ]
        exposed_attrs = ['name'] + self.specific_class_names
        exposed = exposed_methods + exposed_attrs
        if hasattr(self, '_registrars'):
            skip_methods = [
                'scaffold_config',
                'basepath',
            ]

            # Add specific registrar class dict names
            for registrar in self._registrars:
                if isinstance(registrar, SpecificRegistrar):
                    exposed.append(registrar.name)

            # Add general names to main namespace
            register_attrs = [attr for attr in dir(self._general_registrar) if attr not in exposed + skip_methods]
            exposed += register_attrs
        return exposed

    def run(self, section_path_str_or_list: 'RunnerArgs') -> ResultOrResults:
        """
        Use to run registered pipelines/functions/sections. Pass a single section path or a list
        of section paths. If a list is passed, the return value will also be a list, with each
        result item corresponding to the function at the same index.

        If a path for a section is passed rather than a path for a function or pipeline, a list
        of results will be returned for that section as well. Therefore, calling a high-level
        section will result in a nested list structure of results.

        Args:
            section_path_str_or_list: . separated name of path of function or section, or list thereof.
                similar to how a function would be imported. e.g. 'main.data.summarize.summary_func1'
                or when running multiple functions/sections, e.g.
                    ['main.data', 'main.analysis.reg.1']
                These sections/functions are based on the structure of your pipeline_dict

        Returns: result or list of results

        """
        str_or_list_only: Union[str, List[str]] = self._convert_list_or_single_item_view_or_str_to_strs(
            section_path_str_or_list
        )
        additional_items = plugin_manager.plm.hook.pyfileconf_pre_run(
            section_path_str_or_list=section_path_str_or_list, pm=self
        )
        # Will be converted into list always
        str_or_list_only = combine_items_into_list_whether_they_are_lists_or_not_then_extract_from_list_if_only_one_item(
            str_or_list_only, additional_items
        )

        if self.log_folder is not None:
            with stdout_also_logged(self.log_folder):
                return self._run_depending_on_settings(str_or_list_only)
        else:
            return self._run_depending_on_settings(str_or_list_only)

    def run_iter(self, section_path_str_or_list: 'RunnerArgs', config_updates: Sequence[Dict[str, Any]],
                 collect_results: bool = True) -> 'IterativeResults':
        """
        Run one or multiple registered functions/sections multiple times, each time
        updating the config with a combination of the passed config updates.

        Aggregates the updates to each config, then takes the itertools product of all the config
        updates to run the functions with each combination of the configs.

        :param section_path_str_or_list: . separated name of path of function or section, or list thereof.
            similar to how a function would be imported. e.g. 'main.data.summarize.summary_func1'
            or when running multiple functions/sections, e.g. ['main.data', 'main.analysis.reg.1']
        :param config_updates: list of kwarg dictionaries which would normally be provided to .update
        :param collect_results: Whether to aggregate and return results, set to False to save memory
            if results are stored in some other way
        :return:
        """
        iterative_runner = IterativeRunner(
            section_path_str_or_list,
            config_updates,
            base_section_path_str=self.name,
            strip_manager_from_iv=True,
        )
        return iterative_runner.run(collect_results=collect_results)

    def _run_depending_on_settings(self, section_path_str_or_list: Union[str, List[str]]) -> ResultOrResults:
        if self.auto_pdb:
            return self._run_with_auto_pdb(section_path_str_or_list)

        if self.force_continue:
            return self._run_with_force_continue(section_path_str_or_list)

        return self.runner.run(section_path_str_or_list)

    def _run_with_auto_pdb(self, section_path_str_or_list: 'RunnerArgs'):
        pm_func = partial(pdb_post_mortem_or_passed_debug_fn, debug_fn=self.auto_pdb)

        result, successful = _try_except_run_func_except_user_interrupts(
            self.runner.run,
            except_func=pm_func,
            try_func_kwargs=dict(
                section_path_str_or_list=section_path_str_or_list
            ),
        )

        if successful:
            return result
        else:
            return None

    def _run_with_force_continue(self, section_path_str_or_list: Union[str, List[str]]):
        if not isinstance(section_path_str_or_list, list):
            section_path_str_or_list = [section_path_str_or_list]
            strip_list_at_end = True
        else:
            strip_list_at_end = False

        exceptions: List[RunnerException] = []
        results = []
        for section_path_str in section_path_str_or_list:
            result, successful = _try_except_run_func_except_user_interrupts(
                self.runner.run,
                except_func=_return_with_traceback,
                try_func_kwargs=dict(
                    section_path_str_or_list=section_path_str
                ),
            )
            if successful:
                results.append(result)
            else:
                exception, tb = result
                re = RunnerException(
                    exception,
                    section_path_str=section_path_str,
                    trace_back=tb
                )
                exceptions.append(re)
                print(re)

        report_runner_exceptions(exceptions)

        if strip_list_at_end:
            return results[0]
        else:
            return results

    # TODO [#13]: multiple section path strs
    def get(self, section_path_str_or_view: 'StrOrView'):
        section_path_str = self._get_section_path_str_from_section_path_str_or_view(section_path_str_or_view)
        return self.runner.get(section_path_str)

    def reset(self, section_path_str_or_view: 'StrOrView', allow_create: bool = False):
        """
        Resets a function or section config to default.

        To reset all configs, use .reload() instead.

        :param section_path_str_or_view:
        :return:
        """
        section_path_str = self._get_section_path_str_from_section_path_str_or_view(section_path_str_or_view)
        self.config.reset(section_path_str, allow_create=allow_create)

    def reload(self) -> None:
        """
        Useful for getting file system changes without having to start a new Python session. Also resets
        any locally defined configuration.

        Reloads functions from pipeline dict, scaffolds config files for any new functions,
        and updates configuration from files.

        Returns: None

        """
        self._wipe_loaded_modules()
        self.load()

    def load(self) -> None:
        """
        Wrapper to track imported modules so that can reimport them upon reloading
        """
        self._import_tracker = ImportTracker()

        try:
            self._load_pipeline_config_and_runner()
        except Exception as e:
            # Reset loaded modules from import, completely canceling load so it can be tried again
            self._loaded_modules = self._import_tracker.imported_modules
            self._wipe_loaded_modules()
            self._loaded_modules = [] # modules have been wiped, must be sure they won't be wiped again
            raise e

        self._loaded_modules = self._import_tracker.imported_modules

    def update(self, d: dict=None, section_path_str: str=None, **kwargs):
        """
        Update the configuration for an item by section path

        :param d: dictionary of updates
        :param section_path_str: section path of item to be updated
        :param kwargs: kwarg updates
        :return:
        """
        from pyfileconf.selector.models.itemview import _is_item_view

        new_updates_list: List[Dict[str, Any]] = plugin_manager.plm.hook.pyfileconf_pre_update(
            pm=self, d=d, section_path_str=section_path_str, kwargs=kwargs
        )
        new_updates_dict: Dict[str, Any] = {k: v for d in new_updates_list for k, v in d.items()}
        if new_updates_dict:
            if d is not None:
                d.update(new_updates_dict)
            kwargs.update(new_updates_dict)

        self.runner.update(d, section_path_str, **kwargs)

        if section_path_str:
            # If any of the updates are putting an ItemView as the value, then
            # record that this config is dependent on the config referenced
            # by the ItemView
            all_updates = {**kwargs}
            if d is not None:
                all_updates.update(d)
            full_sp = SectionPath.join(self.name, section_path_str)
            for value in all_updates.values():
                if _is_item_view(value):
                    context.add_config_dependency(full_sp, value)

            # Refresh any configs which are dependent on attributes of this config
            # Dependent configs are determined in Selector._get_real_item
            full_sp = SectionPath.join(self.name, section_path_str)
            for dependent_sp in context.force_update_dependencies[full_sp.path_str]:
                manager = self.__class__.get_manager_by_section_path_str(dependent_sp.path_str)
                relative_section_path = SectionPath('.'.join(dependent_sp[1:]))
                manager.refresh(relative_section_path.path_str)

        plugin_manager.plm.hook.pyfileconf_post_update(
            pm=self, d=d, section_path_str=section_path_str, kwargs=kwargs
        )

    def refresh(self, section_path_str: str):
        """
        Reloads from the existing file, then reapplies any config updates. Useful for when
        this config depends on the attribute of some other config which was updated.

        :param section_path_str: section path of item to be refreshed
        :return:
        """
        self.runner.refresh(section_path_str)

    def create(self, section_path_str: str, func_or_class: Optional[Union[Callable, Type]] = None):
        """
        Create a new configuration entry dynamically rather than manually modifying dict file

        :param func_or_class: function or class to use to generate config
        :param section_path_str: section path at which the config should be stored
        :return:
        """
        section_path = SectionPath(section_path_str)

        if section_path[0] in self.specific_class_names:
            # Got path for a specific class dict, add to specific class dict
            if len(section_path) < 3:
                raise ValueError('when targeting a specific class dict, section path must have minimum length of '
                                 '3, e.g. example_class.thing.stuff')
            if func_or_class is not None:
                raise ValueError('only pass func_or_class when targeting main pipeline dict, not specific class dict')
            self._create_specific_class_dict_entry(section_path)
            full_sp = section_path
            registrar = self._registrar_dict[section_path[0]]
            klass = registrar.collection.klass
            if klass is None:
                raise ValueError(f'must have specific class set, got None for class in {registrar}')
            key_attr = registrar.collection.key_attr
            registrar_obj = convert_to_empty_obj_if_necessary(section_path[-1], klass, key_attr=key_attr)
            set_section_path_str = SectionPath.from_section_str_list(section_path[1:-1]).path_str
            scaffold_path_str = SectionPath.from_section_str_list(section_path[1:]).path_str
        else:
            # Got path for main pipeline dict, add to main pipeline dict
            if func_or_class is None:
                raise ValueError('when adding creating item in main pipeline dict, must pass function or class')
            self._create_pipeline_dict_entry(section_path, func_or_class)
            name = func_or_class.__name__
            full_sp = SectionPath.join(section_path, name)
            mod, import_base = get_module_and_name_imported_from(func_or_class)
            obj_import = ObjectImportStatement([name], import_base)
            item = ast.Name(id=name)
            imports = ImportStatementContainer([obj_import])
            registrar_obj = ObjectView.from_ast_and_imports(item, imports)
            if self._general_registrar is None:
                raise ValueError('general registrar must be defined')
            registrar = self._general_registrar
            set_section_path_str = section_path_str
            scaffold_path_str = full_sp.path_str

        registrar.set(set_section_path_str, registrar_obj)
        registrar.scaffold_config_for(scaffold_path_str)
        self.reset(full_sp.path_str, allow_create=True)

    @classmethod
    def get_manager_by_filepath(cls, filepath: str) -> 'PipelineManager':
        """
        Gets the active pipeline manager which references the passed
        filepath.

        :param filepath:
        :return:
        :raises NoPipelineManagerForFilepathException: When no active pipeline
            manager has a config file matching the file path
        """
        abs_filepath = os.path.abspath(filepath)
        for manager in context.active_managers.values():
            manager_path = os.path.abspath(manager.folder)
            if abs_filepath.startswith(manager_path + os.sep):
                return manager

        raise NoPipelineManagerForFilepathException

    @classmethod
    def get_manager_by_section_path_str(cls, section_path_str: str) -> 'PipelineManager':
        """
        Gets the active pipeline manager which references the passed
        section path.

        :param filepath:
        :return:
        :raises NoPipelineManagerForSectionPathException: When no active pipeline
            manager has a config file matching the file path
        """
        section_path = SectionPath(section_path_str)
        manager_name = section_path[0]
        try:
            manager = context.active_managers[manager_name]
            return manager
        except KeyError:
            raise NoPipelineManagerForSectionPathException


    def _create_specific_class_dict_entry(self, section_path: SectionPath):
        name = section_path[0]
        file_path = os.path.join(self.folder, f'{name}_dict.py')
        specific_class_dict_file = SpecificClassDictFile(file_path, name=name + '_dict')
        specific_dict = specific_class_dict_file.load()

        # Modify specific class dict to add entry
        add_item_into_nested_dict_at_section_path(
            specific_dict,
            # skip first section as that is name of specific class, skip last section as
            # that's name which should be created
            section_path[1:-1],
            str(section_path[-1]),  # last section is name that should be created
            add_as_ast_name=False
        )
        specific_dict_str = pretty_format_str('class_dict = ' + str(specific_dict))

        specific_class_dict_config = SpecificClassDictConfig(
            dict(class_dict=specific_dict_str),
            name='specific_class_dict',
            _file=specific_class_dict_file,
        )
        specific_class_dict_file.save(specific_class_dict_config)

    def _create_pipeline_dict_entry(self, section_path: SectionPath, func_or_class: Union[Callable, Type]):
        pipeline_dict_file = PipelineDictFile(self.pipeline_dict_path, name='pipeline_dict')
        pipeline_dict = pipeline_dict_file.load()

        mod, import_base = get_module_and_name_imported_from(func_or_class)
        obj_import = ObjectImportStatement([func_or_class.__name__], import_base)
        imports = ImportStatementContainer([obj_import])

        # Modify pipeline dict to add entry
        add_item_into_nested_dict_at_section_path(pipeline_dict, section_path, func_or_class.__name__)

        # Convert pipeline dict to string for output
        pipeline_dict_str = create_dict_assignment_str_from_nested_dict_with_ast_names(pipeline_dict)

        pipeline_dict_config = PipelineDictConfig(
            dict(pipeline_dict=pipeline_dict_str),
            name='pipeline_dict',
            _file=pipeline_dict_file,
            imports=imports
        )
        pipeline_dict_file.save(pipeline_dict_config)


    def _load_pipeline_config_and_runner(self) -> None:
        """
        External logic for main initialization
        Returns:

        """
        # Load dynamically instead of passing dict to ensure modules are loaded into sys now
        self._registrars, self._general_registrar = create_registrars(
            self.specific_class_config_dicts,
            self.default_config_path,
            self.folder,
            self.pipeline_dict_path,
            manager_name=self.name,
        )

        self.config = ConfigManager(self.default_config_path)
        self.config.load()

        self.runner = Runner(
            config=self.config,
            registrars=self._registrars,
            general_registrar=self._general_registrar,
            name=self.name,
        )

    def _wipe_loaded_modules(self):
        [sys.modules.pop(module) for module in self._loaded_modules]

    def _convert_list_or_single_item_view_or_str_to_strs(self, run_list: 'RunnerArgs') -> Union[str, List[str]]:
        if isinstance(run_list, list):
            return [self._get_section_path_str_from_section_path_str_or_view(run_arg) for run_arg in run_list]
        else:
            return self._get_section_path_str_from_section_path_str_or_view(run_list)

    def _get_section_path_str_from_section_path_str_or_view(self,section_path_str_or_view: 'StrOrView') -> str:
        from pyfileconf.selector.models.itemview import _is_item_view
        if _is_item_view(section_path_str_or_view):
            # ItemView will have PipelineManager.name as first section, must strip
            section_path = SectionPath(section_path_str_or_view.section_path_str)  # type: ignore
            relative_section_path = SectionPath.from_section_str_list(section_path[1:])
            return relative_section_path.path_str
        elif isinstance(section_path_str_or_view, str):
            return section_path_str_or_view
        else:
            raise ValueError(f'expected str or ItemView. Got {section_path_str_or_view} of '
                             f'type {type(section_path_str_or_view)}')

    def _validate_options(self):
        if self.auto_pdb and self.force_continue:
            raise ValueError('cannot force continue and drop into pdb at the same time')
        if self.name in context.active_managers:
            warnings.warn(f'should not repeat names of PipelineManagers. Got repeated name {self.name}. '
                          f'The original PipelineManager has been replaced in the Selector.')

    def _create_project_if_needed(self):
        if self._need_to_create_project():
            create_project(self.folder, self.log_folder, self.specific_class_config_dicts)

    def _need_to_create_project(self) -> bool:
        return any([
            not os.path.exists(self.folder),
            not os.path.exists(self.pipeline_dict_path),
            not os.path.exists(self.default_config_path),
            self.log_folder is not None and not os.path.exists(self.log_folder),
            *[
                not os.path.exists(path) for path in
                [os.path.join(self.folder, f'{name}_dict.py') for name in self.specific_class_names]
            ],
        ])

    def __del__(self):
        try:
            current_manager_under_this_key = context.active_managers[self.name]
        except KeyError:
            # Must have already been deleted from active managers, do nothing
            return

        if current_manager_under_this_key is self:
            del context.active_managers[self.name]

    @property
    def _registrar_dict(self) -> Dict[str, Registrar]:
        if self._general_registrar is None:
            raise ValueError('general registrar must be defined')
        rd: Dict[str, Registrar] = {'general': self._general_registrar}
        if self._registrars is not None:
            for registrar in self._registrars:
                rd[registrar.name] = registrar
        return rd


def _try_except_run_func_except_user_interrupts(try_func: Callable, except_func: Callable = lambda x: x,
                                                exceptions: Sequence[Type[BaseException]] = (
                                                        KeyboardInterrupt,
                                                        SystemExit,
                                                        bdb.BdbQuit
                                                ), try_func_kwargs: Optional[dict] = None,
                                                except_func_kwargs: Optional[dict] = None,
                                                print_traceback: bool = True) -> Tuple[Any, bool]:
    """
    Args:
        try_func:
        except_func: first arg of except_func must accept the exception being raised
        exceptions:
        try_func_kwargs:
        except_func_kwargs:

    Returns:
        Tuple where first item is the result and second item is a boolean for whether an exception was raised.
        In the case of passed exceptions being raised, will quit.
        In the case of another exception being raised, the result of except_func will be returned as the first item
        and False will be the second item of the tuple.

    """
    if try_func_kwargs is None:
        try_func_kwargs = {}

    try:
        return try_func(**try_func_kwargs), True
    except exceptions:  # type: ignore
        quit()
    except Exception as e:
        if print_traceback:
            traceback.print_exc()
        if except_func_kwargs is None:
            except_result = except_func(e)
        else:
            except_result = except_func(e, **except_func_kwargs)
        return except_result, False

def _return_with_traceback(any: Any) -> Tuple[Any, str]:
    return any, traceback.format_exc()

class RunnerException(Exception):

    def __init__(self, *args, section_path_str: 'StrOrView' = None, trace_back: str = None):
        self.section_path_str = section_path_str
        self.trace_back = trace_back

        if len(args) > 0:
            self.exc = args[0]
        else:
            self.exc = None

        super().__init__(*args)

    def __str__(self):
        output_str = f'Error while running {self.section_path_str}:\n\n'
        output_str += self.trace_back

        return output_str

def report_runner_exceptions(runner_exceptions: List[RunnerException]) -> None:
    print('\n\n')
    for runner_exception in runner_exceptions:
        print(runner_exception)
        print('\n\n')

    if len(runner_exceptions) == 0:
        print('Everything ran successfully, no exceptions to report.\n\n')


def create_project(path: str, logs_path: str,
                   specific_class_config_dicts: Optional[List[Dict[str, Union[str, Type, List[str]]]]] = None):
    """
    Creates a new pyfileconf project file structure
    """
    defaults_path = os.path.join(path, 'defaults')
    pipeline_path = os.path.join(path, 'pipeline_dict.py')

    if not os.path.exists(defaults_path):
        os.makedirs(defaults_path)
    if not os.path.exists(logs_path):
        os.makedirs(logs_path)
    if not os.path.exists(pipeline_path):
        with open(pipeline_path, 'w') as f:
            f.write('\npipeline_dict = {}\n')
    if specific_class_config_dicts:
        for specific_class_config in specific_class_config_dicts:
            name = specific_class_config['name']
            dict_path = os.path.join(path, f'{name}_dict.py')
            if not os.path.exists(dict_path):
                with open(dict_path, 'w') as f:
                    f.write(f'\nclass_dict = {{}}\n')



def _validate_registrars(registrars: List[SpecificRegistrar], general_registrar: PipelineRegistrar):
    used_names = dir(general_registrar)
    for registrar in registrars:
        if registrar.name in used_names:
            raise ValueError(f'cannot use a name for a specific class dict which is already specified in '
                             f'top-level pipeline_dict. The issue is with {registrar.name}.')

    all_registrar_names = [registrar.name for registrar in registrars]
    if len(set(all_registrar_names)) != len(all_registrar_names):
        raise ValueError(f'cannot have multiple specific class dicts with the same '
                         f'name attribute. Got names: {all_registrar_names}')


def create_registrars(specific_class_config_dicts: List[SpecificClassConfigDict],
                      basepath: str, pipeline_folder: str, pipeline_dict_path: str,
                      manager_name: Optional[str] = None
                      ) -> Tuple[List[SpecificRegistrar], PipelineRegistrar]:
    registrars, general_registrar = _create_registrars_or_collections_from_dict(
        specific_class_config_dicts,
        basepath,
        pipeline_folder,
        pipeline_dict_path,
        registrar=True,
        manager_name=manager_name,
    )
    registrars = cast(List[SpecificRegistrar], registrars)
    general_registrar = cast(PipelineRegistrar, general_registrar)
    _validate_registrars(registrars, general_registrar)
    for registrar in registrars:
        registrar.scaffold_config()
    general_registrar.scaffold_config()

    return registrars, general_registrar


def create_collections(specific_class_config_dicts: List[SpecificClassConfigDict],
                       basepath: str, pipeline_folder: str, pipeline_dict_path: str,
                       manager_name: Optional[str] = None
                       ) -> Tuple[List[SpecificClassCollection], PipelineCollection]:
    collections, general_collection = _create_registrars_or_collections_from_dict(
        specific_class_config_dicts,
        basepath,
        pipeline_folder,
        pipeline_dict_path,
        registrar=False,
        manager_name=manager_name,
    )
    collections = cast(List[SpecificClassCollection], collections)
    general_collection = cast(SpecificRegistrar, general_collection)
    return collections, general_collection


def _create_registrars_or_collections_from_dict(
    specific_class_config_dicts: List[SpecificClassConfigDict],
    basepath: str, pipeline_folder: str, pipeline_dict_path: str, registrar: bool = True,
    manager_name: Optional[str] = None
) -> Tuple[
        Union[List[SpecificRegistrar], List[SpecificClassCollection]],
        Union[PipelineRegistrar, PipelineCollection]
    ]:
    pipeline_class: Union[Type[PipelineRegistrar], Type[PipelineCollection]]
    specific_class_class: Union[Type[SpecificRegistrar], Type[SpecificClassCollection]]
    if registrar:
        pipeline_class = PipelineRegistrar
        specific_class_class = SpecificRegistrar
    else:
        pipeline_class = PipelineCollection
        specific_class_class = SpecificClassCollection

    # Load dynamically instead of passing dict to ensure modules are loaded into sys now
    pipeline_dict_file = PipelineDictFile(pipeline_dict_path, name='pipeline_dict')
    pipeline_dict = pipeline_dict_file.load()

    objs: Union[List[SpecificRegistrar], List[SpecificClassCollection]] = []
    for specific_class_config_dict in specific_class_config_dicts:

        # Set defaults then update with actual config
        config_dict: Dict[str, Optional[Union[str, Type, List[str]]]] = dict(
            always_assign_strs=None,
            always_import_strs=None,
        )
        config_dict.update(specific_class_config_dict)  # type: ignore

        for key in ('name', 'class'):
            if key not in config_dict or config_dict[key] is None:
                raise ValueError(f'{key} is required in {config_dict}')

        # Kwargs require 'klass' while config dict is 'class'
        kwargs_dict = deepcopy(config_dict)
        kwargs_dict['klass'] = config_dict['class']
        kwargs_dict.pop('class')

        name = cast(str, config_dict['name'])
        file_path = os.path.join(pipeline_folder, f'{name}_dict.py')
        specific_class_dict_file = SpecificClassDictFile(file_path, name=name + '_dict')
        specific_dict = specific_class_dict_file.load()
        obj = specific_class_class.from_dict(
            specific_dict,
            basepath=os.path.join(basepath, name),
            imports=specific_class_dict_file.interface.imports,
            **kwargs_dict  # type: ignore
        )
        objs.append(obj)

    general_obj = pipeline_class.from_dict(
        pipeline_dict,
        basepath=basepath,
        name=manager_name,
        imports=pipeline_dict_file.interface.imports
    )
    return objs, general_obj



import ast
from typing import List, Tuple, Dict, Optional

from pyfileconf.basemodels.container import Container
from pyfileconf.assignments.models.statement import AssignmentStatement
from mixins.repr import ReprMixin

class AssignmentStatementContainer(Container, ReprMixin):
    repr_cols = ['items']

    def __init__(self, items: List[AssignmentStatement]):
        self.items = items

    def __str__(self):
        return ''.join(str(assign) for assign in self)

    def contains_varname(self, varname: str):
        return any([assign.varname == varname for assign in self])

    def to_default_dict_and_annotation_dict(self) -> Tuple[Dict[str, ast.AST], Dict[str, ast.Name]]:
        default_dict: Dict[str, ast.AST] = {}
        annotation_dict: Dict[str, ast.Name] = {}
        for assign_statement in self.items:
            default_item, annotation_item = assign_statement.to_default_dict_and_annotation_dict()
            default_dict.update(default_item)
            annotation_dict.update(annotation_item)

        return default_dict, annotation_dict

    @classmethod
    def from_dict_of_varnames_and_ast(cls, assignment_dict: Dict[str, ast.AST],
                                      annotation_dict: Dict[str, ast.Name] = None,
                                      preferred_position: str = None):
        if annotation_dict is None:
            annotation_dict = {}

        assigns = []
        for varname, value in assignment_dict.items():

            # Extract annotation if exists
            annotation: Optional[ast.Name]
            if varname in annotation_dict:
                annotation = annotation_dict[varname]
            else:
                annotation = None

            # Create assignment statement
            assigns.append(
                AssignmentStatement.from_varname_and_ast_value(
                    varname,
                    value=value,
                    annotation=annotation,
                    preferred_position=preferred_position
                )
            )

        return cls(assigns)



import os
from copy import deepcopy
from typing import Optional, Sequence

from pyfileconf import Selector
from pyfileconf.main import create_project
from tests.test_pipeline_manager.base import BASE_GENERATED_DIR, CLASS_CONFIG_DICT_LIST, PipelineManagerTestBase, \
    FULL_CLASS_DICT_LIST, SAME_CLASS_CONFIG_DICT_LIST, DIFFERENT_CLASS_CONFIG_DICT_LIST
from tests.utils import delete_project


def _assert_project_has_correct_files(folder: str):
    defaults_path = os.path.join(folder, 'defaults')
    pipeline_folder = folder
    pipeline_dict_path = os.path.join(pipeline_folder, 'pipeline_dict.py')
    example_class_dict_path = os.path.join(folder, 'example_class_dict.py')
    logs_path = os.path.join(folder, 'MyLogs')
    all_paths = [
        defaults_path,
        pipeline_dict_path,
        example_class_dict_path,
        logs_path
    ]
    for path in all_paths:
        assert os.path.exists(path)

    with open(pipeline_dict_path, 'r') as f:
        contents = f.read()
        assert 'pipeline_dict = {}' in contents

    with open(example_class_dict_path, 'r') as f:
        contents = f.read()
        assert 'class_dict = {}' in contents


def test_create_project():
    logs_path = os.path.join(BASE_GENERATED_DIR, 'MyLogs')
    delete_project(BASE_GENERATED_DIR, logs_path, CLASS_CONFIG_DICT_LIST)
    create_project(BASE_GENERATED_DIR, logs_path, CLASS_CONFIG_DICT_LIST)

    _assert_project_has_correct_files(BASE_GENERATED_DIR)

    delete_project(BASE_GENERATED_DIR, logs_path, CLASS_CONFIG_DICT_LIST)


class PipelineManagerLoadTestBase(PipelineManagerTestBase):

    def assert_default_imports_and_assigns_in_contents(self, contents: str):
        assert 'from pyfileconf import Selector' in contents
        assert 's = Selector()' in contents

    def assert_a_function_config_file_contents(self, contents: str):
        self.assert_default_imports_and_assigns_in_contents(contents)
        assert 'from typing import List' in contents
        assert 'from tests.input_files.mypackage.cmodule import ExampleClass' in contents
        assert 'a: ExampleClass = None' in contents
        assert 'b: List[str] = None' in contents

    def assert_example_class_config_file_contents(self, contents: str):
        self.assert_default_imports_and_assigns_in_contents(contents)
        assert 'import typing' in contents
        assert 'from typing import Optional' in contents
        assert 'a: typing.Tuple[int, int] = None' in contents
        assert 'name: Optional[str] = None' in contents

    def assert_example_class_dict_config_file_contents(self, contents: str, imports: Optional[Sequence[str]] = None,
                                                       assigns: Optional[Sequence[str]] = None,
                                                       a_value: str = 'a: typing.Tuple[int, int] = None',
                                                       name_value: str = "name: Optional[str] = 'data'"):
        if imports is None:
            imports = [
                'from pyfileconf import Selector',
            ]

        if assigns is None:
            assigns = [
                's = Selector()',
            ]

        # Always imports, from typing of arguments
        assert "from typing import Optional" in contents
        assert "import typing" in contents
        assert "import collections" in contents
        assert "from traceback import TracebackException" in contents

        # Always assigns, from arguments
        assert "c: Optional[collections.defaultdict] = None" in contents
        assert "d: Optional['TracebackException'] = None" in contents

        # TODO [#67]: format config output better, then update this test assertion
        #
        # For some reason, longer items when written to the config are line breaking at weird parts
        assert "f: typing.Sequence[typing.Union[collections.Counter, pathlib.Path]] " \
               "= (\n    collections.Counter(), collections.Counter(), pathlib.Path('.'))" in contents

        for item in imports + assigns + [a_value, name_value]:
            assert item in contents

    def assert_second_example_class_dict_config_file_contents(self, contents: str):
        assert "from typing import Optional" in contents
        assert "from tests.input_files.mypackage.cmodule import ExampleClass" in contents
        assert "s = Selector()" in contents
        assert "b: ExampleClass = None" in contents
        assert "name: Optional[str] = 'data'" in contents


class TestPipelineManagerLoad(PipelineManagerLoadTestBase):

    def test_create_empty_pm(self):
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager

    def test_create_project_with_pm(self):
        delete_project(BASE_GENERATED_DIR, self.logs_path, FULL_CLASS_DICT_LIST)
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=FULL_CLASS_DICT_LIST
        )
        pipeline_manager.load()
        _assert_project_has_correct_files(self.pm_folder)

    def test_create_pm_with_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        module_folder = os.path.join(self.defaults_path, 'stuff')
        function_path = os.path.join(module_folder, 'a_function.py')
        with open(function_path, 'r') as f:
            contents = f.read()
            self.assert_a_function_config_file_contents(contents)

    def test_create_multiple_pms_with_function(self):
        self.write_a_function_to_pipeline_dict_file()
        self.write_a_function_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()

        # Assert pipeline manager 1 contents
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        module_folder = os.path.join(self.defaults_path, 'stuff')
        function_path = os.path.join(module_folder, 'a_function.py')
        with open(function_path, 'r') as f:
            contents = f.read()
            self.assert_a_function_config_file_contents(contents)

        # Assert pipeline manager 2 contents
        sel = Selector()
        iv = sel.test_pipeline_manager2.stuff.a_function
        module_folder = os.path.join(self.second_defaults_path, 'stuff')
        function_path = os.path.join(module_folder, 'a_function.py')
        with open(function_path, 'r') as f:
            contents = f.read()
            self.assert_a_function_config_file_contents(contents)

    def test_create_pm_with_class(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.ExampleClass
        module_folder = os.path.join(self.defaults_path, 'stuff')
        class_path = os.path.join(module_folder, 'ExampleClass.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_config_file_contents(contents)

    def test_create_multiple_pms_with_class(self):
        self.write_example_class_to_pipeline_dict_file()
        self.write_example_class_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()

        # Assert pipeline manager 1 contents
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.ExampleClass
        module_folder = os.path.join(self.defaults_path, 'stuff')
        class_path = os.path.join(module_folder, 'ExampleClass.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_config_file_contents(contents)

        # Assert pipeline manager 2 contents
        sel = Selector()
        iv = sel.test_pipeline_manager2.stuff.ExampleClass
        module_folder = os.path.join(self.second_defaults_path, 'stuff')
        class_path = os.path.join(module_folder, 'ExampleClass.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_config_file_contents(contents)

    def test_create_pm_with_class_dict(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'stuff')
        class_path = os.path.join(module_folder, 'data.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(contents)

    def test_create_multiple_pms_with_class_dict(self):
        self.write_example_class_dict_to_file()
        self.write_example_class_dict_to_file(pm_index=1)
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager2.load()
        sel = Selector()

        # Assert pipeline manager 1 contents
        iv = sel.test_pipeline_manager.example_class.stuff.data
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'stuff')
        class_path = os.path.join(module_folder, 'data.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(contents)

        # Assert pipeline manager 2 contents
        iv = sel.test_pipeline_manager2.example_class.stuff.data
        class_folder = os.path.join(self.second_defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'stuff')
        class_path = os.path.join(module_folder, 'data.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(contents)

    def test_create_pm_with_class_dict_and_custom_key_attr(self):
        self.write_example_class_dict_to_file()
        class_config_dict_list = deepcopy(CLASS_CONFIG_DICT_LIST)
        class_config_dict_list[0].update(
            key_attr='a'
        )
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=class_config_dict_list
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'stuff')
        class_path = os.path.join(module_folder, 'data.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(
                contents,
                a_value="a: typing.Tuple[int, int] = 'data'",
                name_value="name: Optional[str] = None"
            )

    def test_create_pm_with_class_dict_and_imports(self):
        self.write_example_class_dict_to_file()
        always_imports = [
            'from copy import deepcopy',
            'from functools import partial'
        ]
        always_assigns = [

        ]
        class_config_dict_list = deepcopy(CLASS_CONFIG_DICT_LIST)
        class_config_dict_list[0].update(
            always_import_strs=always_imports,
            always_assign_strs=always_assigns,
        )
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=class_config_dict_list,
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'stuff')
        class_path = os.path.join(module_folder, 'data.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(
                contents,
                imports=always_imports,
                assigns=always_assigns
            )

    def test_create_pm_with_class_dict_and_assigns(self):
        self.write_example_class_dict_to_file()
        always_imports = [

        ]
        always_assigns = [
            'my_var = 6',
            'stuff = list((1,))'
        ]
        class_config_dict_list = deepcopy(CLASS_CONFIG_DICT_LIST)
        class_config_dict_list[0].update(
            always_import_strs=always_imports,
            always_assign_strs=always_assigns,
        )
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=class_config_dict_list,
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'stuff')
        class_path = os.path.join(module_folder, 'data.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(
                contents,
                imports=always_imports,
                assigns=always_assigns
            )

    def test_create_pm_with_class_dict_imports_and_assigns(self):
        self.write_example_class_dict_to_file()
        always_imports = [
            'from copy import deepcopy',
            'from functools import partial'
        ]
        always_assigns = [
            'my_var = 6',
            'stuff = list((1,))'
        ]
        class_config_dict_list = deepcopy(CLASS_CONFIG_DICT_LIST)
        class_config_dict_list[0].update(
            always_import_strs=always_imports,
            always_assign_strs=always_assigns,
        )
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=class_config_dict_list,
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        class_folder = os.path.join(self.defaults_path, 'example_class')
        module_folder = os.path.join(class_folder, 'stuff')
        class_path = os.path.join(module_folder, 'data.py')
        with open(class_path, 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(
                contents,
                imports=always_imports,
                assigns=always_assigns
            )

    def test_create_pm_with_multiple_class_dicts_same_class(self):
        self.write_example_class_dict_to_file()  # example_class
        self.write_example_class_dict_to_file(1)  # example_class2
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=SAME_CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        class_folders = [
            os.path.join(self.defaults_path, 'example_class'),
            os.path.join(self.defaults_path, 'example_class2'),
        ]
        module_folders = [os.path.join(class_folder, 'stuff') for class_folder in class_folders]
        class_paths = [os.path.join(module_folder, 'data.py') for module_folder in module_folders]
        for class_path in class_paths:
            with open(class_path, 'r') as f:
                contents = f.read()
                self.assert_example_class_dict_config_file_contents(contents)

    def test_create_pm_with_multiple_class_dicts_different_class(self):
        self.write_example_class_dict_to_file()  # example_class
        self.write_example_class_dict_to_file(2)  # second_example_class
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=DIFFERENT_CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        class_folders = [
            os.path.join(self.defaults_path, 'example_class'),
            os.path.join(self.defaults_path, 'second_example_class'),
        ]
        module_folders = [os.path.join(class_folder, 'stuff') for class_folder in class_folders]
        class_paths = [os.path.join(module_folder, 'data.py') for module_folder in module_folders]
        with open(class_paths[0], 'r') as f:
            contents = f.read()
            self.assert_example_class_dict_config_file_contents(contents)
        with open(class_paths[1], 'r') as f:
            contents = f.read()
            self.assert_second_example_class_dict_config_file_contents(contents)


class TestPipelineManagerInvalidLoad(PipelineManagerLoadTestBase):

    def test_create_pm_with_specific_class_name_matching_pipeline_name(self):
        self.write_a_function_to_pipeline_dict_file()
        self.write_example_class_dict_to_file()
        specific_class_config = deepcopy(CLASS_CONFIG_DICT_LIST)
        specific_class_config[0].update(name='stuff')  # make name match section in main dict
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=specific_class_config
        )
        with self.assertRaises(ValueError) as cm:
            pipeline_manager.load()
            exc = cm.exception
            assert 'cannot use a name for a specific class dict which is already specified ' \
                   'in top-level pipeline_dict' in str(exc)

    def test_create_pm_with_two_identical_specific_class_dicts(self):
        self.write_example_class_dict_to_file()  # example_class
        self.write_example_class_dict_to_file(1)  # example_class2
        specific_class_config = deepcopy(SAME_CLASS_CONFIG_DICT_LIST)
        # make name in second specific dict match that of first
        specific_class_config[1].update(name=specific_class_config[0]['name'])
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=specific_class_config,
        )
        with self.assertRaises(ValueError) as cm:
            pipeline_manager.load()
            exc = cm.exception
            assert 'cannot have multiple specific class dicts with the same name attribute. Got names' in str(exc)



import itertools
from collections import defaultdict
from copy import deepcopy
from typing import List, Dict, Any, Tuple, Sequence, Optional

from pyfileconf.plugin import manager
from pyfileconf.runner.models.interfaces import RunnerArgs, IterativeResults
from pyfileconf.sectionpath.sectionpath import SectionPath


class IterativeRunner:
    """
    Class that enables running functions/sections with different combinations of config updates,
    even across multiple pipeline managers

    Run one or multiple registered functions/sections multiple times, each time
    updating the config with a combination of the passed config updates.

    Aggregates the updates to each config, then takes the itertools product of all the config
    updates to run the functions with each combination of the configs.
    """

    run_items: List[SectionPath]
    config_updates: Sequence[Dict[str, Any]]
    cases: List[Tuple[Dict[str, Any], ...]]

    def __init__(
        self,
        section_path_str_or_list: RunnerArgs,
        config_updates: Sequence[Dict[str, Any]],
        base_section_path_str: Optional[str] = None,
        strip_manager_from_iv: bool = False,
    ):
        """

        :param section_path_str_or_list: . separated name of path of function or section, or list thereof.
            similar to how a function would be imported. e.g. 'main.data.summarize.summary_func1'
            or when running multiple functions/sections, e.g. ['main.data', 'main.analysis.reg.1']
        :param config_updates: list of kwarg dictionaries which would normally be provided to .update
        :param base_section_path_str: section path str to put at beginning of all passed section paths
        :param strip_manager_from_iv: whether to remove manager name from any incoming item views
        """
        self.base_section_path_str = base_section_path_str
        self.run_items = SectionPath.list_from_ambiguous(
            section_path_str_or_list,
            base_section_path_str=base_section_path_str,
            strip_manager_from_iv=strip_manager_from_iv,
        )
        self.config_updates = config_updates
        self.cases = self.get_cases()

    def get_cases(self) -> List[Tuple[Dict[str, Any], ...]]:
        cases_lol: List[
            List[Tuple[Dict[str, Any], ...]]
        ] = manager.plm.hook.pyfileconf_iter_get_cases(config_updates=self.config_updates, runner=self)
        cases = list(itertools.chain(*cases_lol))
        manager.plm.hook.pyfileconf_iter_modify_cases(cases=cases, runner=self)
        return cases

    def run(self, collect_results: bool = True) -> IterativeResults:
        """
        :param collect_results: Whether to aggregate and return results, set to False to save memory
            if results are stored in some other way
        :return:
        """
        from pyfileconf.main import PipelineManager

        all_results = []
        for case in self.cases:
            for config_dict in case:
                sp_str = config_dict["section_path_str"]
                if self.base_section_path_str is not None:
                    sp_str = SectionPath.join(
                        self.base_section_path_str, sp_str
                    ).path_str
                manager = PipelineManager.get_manager_by_section_path_str(sp_str)
                relative_section_path_str = SectionPath(
                    ".".join(SectionPath(sp_str)[1:])
                ).path_str
                manager.reset(relative_section_path_str)
                relative_conf_dict = deepcopy(config_dict)
                relative_conf_dict["section_path_str"] = relative_section_path_str
                manager.update(**relative_conf_dict)
            result = self._run()
            if collect_results:
                in_out_tup = (case, result)
                all_results.append(in_out_tup)
        return all_results

    def _run(self) -> Any:
        from pyfileconf.main import PipelineManager

        results = []
        for sp in self.run_items:
            # Look up appropriate manager and run it
            manager = PipelineManager.get_manager_by_section_path_str(sp.path_str)
            relative_section_path_str = SectionPath(".".join(sp[1:])).path_str
            result = manager.run(relative_section_path_str)
            results.append(result)

        if len(results) == 1:
            return results[0]

        return results


def get_config_product(
    config_dicts: Sequence[Dict[str, Any]]
) -> List[Tuple[Dict[str, Any], ...]]:
    """

    :param config_dicts:
    :return:

    :Examples:

        >>> cd = dict(
        ...     section_path_str='abc',
        ...     a=10
        ... )
        >>> cd2 = dict(
        ...     section_path_str='abc',
        ...     a=20
        ... )
        >>> cd3 = dict(
        ...     section_path_str='def',
        ...     a=20
        ... )
        >>> cds = [cd, cd2, cd3]
        >>> result = get_config_product(cds)
        >>> result
        ... [({'section_path_str': 'abc', 'a': 10}, {'section_path_str': 'def', 'a': 20}),
        ... ({'section_path_str': 'abc', 'a': 20}, {'section_path_str': 'def', 'a': 20})]
    """
    by_section_path_dict = _get_configs_by_section_path_dict(config_dicts)
    return list(itertools.product(*by_section_path_dict.values()))


def _get_configs_by_section_path_dict(
    config_dicts: Sequence[Dict[str, Any]]
) -> Dict[str, List[Dict[str, Any]]]:
    """

    :param config_dicts:
    :return:

    :Examples:

        >>> cd = dict(
        ...     section_path_str='abc',
        ...     a=10
        ... )
        >>> cd2 = dict(
        ...     section_path_str='abc',
        ...     a=20
        ... )
        >>> cd3 = dict(
        ...     section_path_str='def',
        ...     a=20
        ... )
        >>> cds = [cd, cd2, cd3]
        >>> result = _get_configs_by_section_path_dict(cds)
        >>> result
        ... {'abc': [{'section_path_str': 'abc', 'a': 10},
        ...       {'section_path_str': 'abc', 'a': 20}],
        ...      'def': [{'section_path_str': 'def', 'a': 20}]})
    """
    combined_dicts: Dict[str, List[Dict[str, Any]]] = defaultdict(lambda: [])
    for config_dict in config_dicts:
        conf = deepcopy(config_dict)
        sp = conf["section_path_str"]
        combined_dicts[sp].append(conf)
    return combined_dicts



from typing import Union, cast
import ast

AstDict = Union[ast.Dict, ast.Call]

def extract_dict_from_ast_dict_or_dict_constructor(ast_dict: AstDict) -> dict:
    if isinstance(ast_dict, ast.Dict):
        return _extract_dict_from_ast_dict(ast_dict)
    elif isinstance(ast_dict, ast.Call):
        return _extract_dict_from_ast_dict_constructor(ast_dict)
    else:
        raise ValueError(f'expected ast.Dict or ast.Call, got {ast_dict} of type {type(ast_dict)}')

def _extract_dict_from_ast_dict(ast_dict: ast.Dict) -> dict:
    out_dict = {}
    for ast_key, ast_value in zip(ast_dict.keys, ast_dict.values):
        ast_key = cast(ast.Str, ast_key)
        # TODO [#25]: remove type ignores from _extract_dict_from_ast_dict when ast has better typing support
        key = ast_key.s  # type: ignore
        out_dict[key] = ast_value
    return out_dict

def _extract_dict_from_ast_dict_constructor(ast_dict: ast.Call) -> dict:
    from pyfileconf.io.file.load.parsers.kwargs import extract_keywords_from_ast
    ast_kwargs = extract_keywords_from_ast(ast_dict)
    return {key: ast_value for key, ast_value in ast_kwargs.items()}



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pyfileconf.config.models.config import ActiveFunctionConfig

from pyfileconf.pipelines.models.file import FunctionConfigFile
from pyfileconf.io.file.interfaces.activeconfig import ActiveConfigFileInterface

class ActiveFunctionConfigFile(FunctionConfigFile):
    # class to use for interfacing with file
    # no need to override default
    interface_class = ActiveConfigFileInterface


    def load(self, config_class: type = None) -> 'ActiveFunctionConfig':
        # Override base class method to pull a single dict, and not pass annotations
        from pyfileconf.config.models.config import ActiveFunctionConfig
        user_defined_dict = self.interface.load()

        if config_class is None:
            config_class = ActiveFunctionConfig

        return config_class(
            d=user_defined_dict,
            imports=self.interface.imports,
            _file=self,
            name=self.name
        )


from typing import TYPE_CHECKING, Optional, Type, Sequence

from pyfileconf.exceptions.imports import ExtractedIncorrectTypeOfImportException
from pyfileconf.imports.models.statements.module import ModuleImportStatement

if TYPE_CHECKING:
    from pyfileconf.data.models.config import SpecificClassConfig

from pyfileconf.basemodels.file import ConfigFileBase
from pyfileconf.imports.models.statements.obj import ObjectImportStatement
from pyfileconf.assignments.models.statement import AssignmentStatement
from pyfileconf.io.file.interfaces.activeconfig import ActiveConfigFileInterface


class SpecificClassConfigFile(ConfigFileBase):
    # lines to always import. pass import objects
    always_imports = [
        ObjectImportStatement.from_str('from pyfileconf import Selector', preferred_position='begin')
    ]

    # assignment lines to always include at beginning. pass assign objects
    always_assigns = [
        AssignmentStatement.from_str('s = Selector()', preferred_position='begin'),
    ]

    def __init__(self, filepath: str, name: str = None, klass: Optional[Type] = None,
                 always_import_strs: Optional[Sequence[str]] = None,
                 always_assign_strs: Optional[Sequence[str]] = None):
        super().__init__(
            filepath,
            name=name,
            klass=klass,
            always_import_strs=always_import_strs,
            always_assign_strs=always_assign_strs
        )
        # Override class definitions with object specific definitions, if specifics are passed
        if self.always_import_strs:
            imports = []
            for import_str in self.always_import_strs:
                try:
                    imports.append(ObjectImportStatement.from_str(import_str))
                except ExtractedIncorrectTypeOfImportException:
                    imports.append(ModuleImportStatement.from_str(import_str))
            self.always_imports = imports
        elif self.always_import_strs == []:
            # None passed, remove default imports
            self.always_imports = []

        if self.always_assign_strs:
            self.always_assigns = [AssignmentStatement.from_str(assign_str) for assign_str in self.always_assign_strs]
        elif self.always_assign_strs == []:
            # None passed, remove default assignments
            self.always_assigns = []


    def __call__(self, *args, **kwargs):
        """
        For compatibility with BaseConfig which expects to call class, while here an object will be used
        """
        # Create new object
        # Use defaults from this object
        obj_kwargs = dict(
            name=self.name,
            klass=self.klass,
            always_import_strs=self.always_import_strs,
            always_assign_strs=self.always_assign_strs
        )
        obj_kwargs.update(kwargs)
        obj = self.__class__(*args, **obj_kwargs)
        return obj

    def load(self, config_class: type = None) -> 'SpecificClassConfig':
        # Override base class method to pull a single dict, and not pass annotations
        from pyfileconf.data.models.config import SpecificClassConfig

        config_dict, annotation_dict = self.interface.load()

        if config_class is None:
            config_class = SpecificClassConfig

        return config_class(
            d=config_dict,
            annotations=annotation_dict,
            imports=self.interface.imports,
            _file=self,
            name=self.name,
            klass=self.klass,
            always_assign_strs=self.always_assign_strs,
            always_import_strs=self.always_import_strs,
        )


import inspect
from typing import Callable, Tuple, cast, Sequence, Type, Union, Dict, Any, List
from functools import partial

from mixins.repr import ReprMixin

from pyfileconf.basemodels.collection import Collection
from pyfileconf.basemodels.registrar import Registrar
from pyfileconf.config.models.manager import ConfigManager, ActiveFunctionConfig
from pyfileconf.data.models.collection import SpecificClassCollection
from pyfileconf.exceptions.config import ConfigManagerNotLoadedException
from pyfileconf.exceptions.registrar import NoRegistrarWithNameException
from pyfileconf.logic.combine import \
    combine_items_into_list_whether_they_are_lists_or_not_then_extract_from_list_if_only_one_item
from pyfileconf.pipelines.models.registrar import PipelineRegistrar, PipelineCollection
from pyfileconf.logic.get import _get_public_name_or_special_name
from pyfileconf.plugin import manager
from pyfileconf.runner.models.interfaces import (
    StrOrListOfStrs,
    Result,
    Results,
    ResultOrResults
)
from pyfileconf.pipelines.models.interfaces import (
    PipelineOrFunctionOrCollection,
    PipelineOrFunction
)
from pyfileconf.sectionpath.sectionpath import SectionPath
from pyfileconf.config.logic.write import dict_as_function_kwarg_str
from pyfileconf.basemodels.pipeline import Pipeline
from pyfileconf.tracing import RunningTracker
from pyfileconf.views.object import ObjectView

class Runner(ReprMixin):
    repr_cols = ['_config', '_pipelines']

    def __init__(self, config: ConfigManager, registrars: Sequence[Registrar], general_registrar: PipelineRegistrar,
                 name: str):
        self._config = config
        self._registrars = registrars
        self._general_registrar = general_registrar
        self._manager_name = name

        self._full_getattr = ''
        self._all_specific_classes = tuple([registrar.klass for registrar in self._registrars])
        self._specific_class_registrar_map = {registrar.klass: registrar for registrar in self._registrars}
        self._loaded_objects: Dict[str, Any] = {}

    def __getattr__(self, item):
        # TODO [#14]: find way of doing runner look ups with fewer side effects
        # TODO [#15]: find a way to get __getattr__ logic for runner working for sections
        self._full_getattr += item
        try:
            func_or_collection = self._get_func_or_collection(self._full_getattr)
        except KeyError as e:
            self._full_getattr = '' # didn't find an item, must be incorrect path. reset total path
            raise e

        if isinstance(func_or_collection, (PipelineCollection, SpecificClassCollection)):
            # Got section, need to keep going. Return self
            self._full_getattr += '.' # add period to separate next section
            return self
        else:
            # Got function or Pipeline, return the function or pipeline itself
            full_getattr = self._full_getattr
            self._full_getattr = ''  # found the item, reset for next time
            configured_func_or_pipeline = self.get(full_getattr)
            return configured_func_or_pipeline

    # TODO [#16]: this may work once __getattr__ works with sections
    # def __dir__(self):
    #     exposed_methods = [
    #         'run',
    #         'get',
    #     ]
    #
    #     if self._full_getattr != '':
    #         # Already into nested runner, need to pull appropriate collection
    #         try:
    #             func_or_collection = self._get_func_or_collection(self._full_getattr)
    #             pipeline_attrs = [attr for attr in dir(func_or_collection) if attr not in exposed_methods]
    #         except KeyError:
    #             pipeline_attrs = []
    #
    #     else:
    #         # Original runner. Just pull pipeline register attrs
    #         pipeline_attrs = [attr for attr in dir(self._pipelines) if attr not in exposed_methods]
    #
    #     return exposed_methods + pipeline_attrs


    def run(self, section_path_str_or_list: StrOrListOfStrs) -> ResultOrResults:
        """
        Use to run registered pipelines/functions/sections. Pass a single section path or a list
        of section paths. If a list is passed, the return value will also be a list, with each
        result item corresponding to the function at the same index.

        If a path for a section is passed rather than a path for a function or pipeline, a list
        of results will be returned for that section as well. Therefore, calling a high-level
        section will result in a nested list structure of results.

        Args:
            section_path_str_or_list: . separated name of path of function or section, or list thereof.
                similar to how a function would be imported. e.g. 'main.data.summarize.summary_func1'
                or when running multiple functions/sections, e.g.
                    ['main.data', 'main.analysis.reg.1']
                These sections/functions are based on the structure of your pipeline_dict

        Returns: result or list of results

        """
        multiple_results = False
        if isinstance(section_path_str_or_list, str):
            # Running single function/pipeline
            result = self._run(section_path_str_or_list)
        elif isinstance(section_path_str_or_list, list):
            multiple_results = True
            result = [self._run(section_path_str) for section_path_str in section_path_str_or_list]
        else:
            raise ValueError('must pass str or list of strs of section paths to Runner.run')

        additional_results = manager.plm.hook.pyfileconf_post_run(results=result, runner=self)
        if additional_results and multiple_results:
            result.extend(additional_results)
        elif additional_results and not multiple_results:
            result = [result, *additional_results]

        return result

    def _run(self, section_path_str: str) -> ResultOrResults:
        """
        Internal run function for running a single section path string. Handles both running
        sections and running individual functions/pipelines

        Args:
            section_path_str:

        Returns:

        """
        func_or_collection = self._get_func_or_collection(section_path_str)
        if isinstance(func_or_collection, PipelineCollection):
            return self._run_section(section_path_str)
        elif type(func_or_collection) is type and issubclass(type(func_or_collection), Pipeline):
            return self._run_one_pipeline(section_path_str)
        elif self._is_specific_class(func_or_collection):
            return self._run_one_specific_class(section_path_str)
        elif inspect.isclass(func_or_collection):
            return self._run_one_class(section_path_str)
        elif callable(func_or_collection):
            return self._run_one_func(section_path_str)
        else:
            raise ValueError(f'could not run section {section_path_str}. expected PipelineCollection or function,'
                             f'got {func_or_collection} of type {type(func_or_collection)}')


    def _run_section(self, section_path_str: str) -> Results:
        section = self._get_func_or_collection(section_path_str)
        section = cast(PipelineCollection, section)
        results = []
        for section_or_object_view in section:

            # Get section path by which to call this item
            subsection_name = _get_public_name_or_special_name(section_or_object_view, accept_output_names=False)
            subsection_path_str = SectionPath.join(section_path_str, subsection_name).path_str

            # Get from object view if necessary
            if isinstance(section_or_object_view, ObjectView):
                section_or_callable = section_or_object_view.item
            else:
                section_or_callable = section_or_object_view

            if isinstance(section_or_callable, PipelineCollection):
                # got another section within this section. recursively call run section
                results.append(self._run_section(subsection_path_str))
            elif type(section_or_callable) is type and issubclass(section_or_callable, Pipeline):
                # run pipeline
                results.append(self._run_one_pipeline(subsection_path_str))
            elif self._is_specific_class(section_or_callable):
                results.append(self._run_one_specific_class(section_path_str))
            elif inspect.isclass(section_or_callable):
                results.append(self._run_one_class(section_path_str))
            elif callable(section_or_callable):
                # run function
                results.append(self._run_one_func(subsection_path_str))
            else:
                raise ValueError(f'could not run section {subsection_path_str}. expected PipelineCollection or '
                                 f'function or Pipeline,'
                                 f'got {section_or_callable} of type {type(section_or_callable)}')

        return results


    def _run_one_func(self, section_path_str: str) -> Result:
        with RunningTracker(section_path_str, base_section_path_str=self._manager_name):
            func, config_dict = self._get_func_and_config(section_path_str)

            print(f'Running function {section_path_str}({dict_as_function_kwarg_str(config_dict)})')
            result = func(**config_dict)
            print(f'Result:\n{result}\n')
        self._add_to_config_dependencies_if_necessary(section_path_str)
        return result

    def _run_one_class(self, section_path_str: str) -> Result:
        with RunningTracker(section_path_str, base_section_path_str=self._manager_name):
            klass, config_dict = self._get_func_and_config(section_path_str)
            obj = klass(**config_dict)

            print(f'Running class {section_path_str}({dict_as_function_kwarg_str(config_dict)})')
            result = obj()
            print(f'Result:\n{result}\n')
        self._add_to_config_dependencies_if_necessary(section_path_str)
        return result

    def _run_one_specific_class(self, section_path_str: str) -> Result:
        with RunningTracker(section_path_str, base_section_path_str=self._manager_name):
            klass, config_dict = self._get_class_and_config(section_path_str)
            obj = klass(**config_dict)
            self._add_full_section_path_str_to_obj(section_path_str, obj)
            registrar = self._specific_class_registrar_map[klass]
            execute_attr = registrar.execute_attr
            func = getattr(obj, execute_attr)

            print(f'Running class {section_path_str}({dict_as_function_kwarg_str(config_dict)})')
            result = func()
            print(f'Result:\n{result}\n')
        self._add_to_config_dependencies_if_necessary(section_path_str)
        return result

    def _run_one_pipeline(self, section_path_str: str) -> Result:
        with RunningTracker(section_path_str, base_section_path_str=self._manager_name):
            pipeline_class, config_dict = self._get_pipeline_and_config(section_path_str)

            # Construct new pipeline instance with config args
            configured_pipeline = pipeline_class(**config_dict)  # type: ignore

            print(f'Running pipeline {configured_pipeline}({dict_as_function_kwarg_str(config_dict)})')
            result = configured_pipeline.execute()
            print(f'Result:\n{result}\n')
        self._add_to_config_dependencies_if_necessary(section_path_str)
        return configured_pipeline

    def get(self, section_path_str: str):
        func_or_collection = self._get_func_or_collection(section_path_str)
        if isinstance(func_or_collection, Collection):
            return self._get_section(section_path_str)
        elif type(func_or_collection) is type and issubclass(type(func_or_collection), Pipeline):
            # Got pipeline class
            return self._get_one_pipeline_with_config(section_path_str)
        elif self._is_specific_class(func_or_collection):
            return self._get_one_obj_with_config(section_path_str)
        elif callable(func_or_collection):
            return self._get_one_func_with_config(section_path_str)
        else:
            raise ValueError(f'could not get section {section_path_str}. expected PipelineCollection or function,'
                             f'got {func_or_collection} of type {type(func_or_collection)}')

    # TODO [#50]: restructure runner get and run
    #
    # Currently the checks for section, function, specific class are very repetitive across the
    # various get and run functions. Standardize these and offload into a single function.

    def _get_section(self, section_path_str: str):
        section = self._get_func_or_collection(section_path_str)
        section = cast(Collection, section)

        # Need to handle definition structure which can be lists inside dicts or more dicts inside dicts
        # This method will be called recursively on each section. Check to see if there are any sections
        # inside this section. If so, then that section would be defined by a dict key in the definition,
        # and therefore results should be put in a dict. If there are no sections within this section, then
        # a list was used to store the items in this section and so results will be put in a list.
        results: Union[Dict[str, Union[Collection, Any]], List[Any]]
        if any(isinstance(item, Collection) for item in section):
            results = {}
        else:
            results = []

        for section_or_object_view in section:

            # Get from object view if necessary
            if isinstance(section_or_object_view, ObjectView):
                section_or_callable = section_or_object_view.item
            else:
                section_or_callable = section_or_object_view

            # Get section path by which to call this item
            if self._is_specific_class(section_or_callable):
                # If specific class, need to look up which key holds the name
                item_registrar = self._specific_class_registrar_map[type(section_or_callable)]
                item_key_attr = item_registrar.key_attr
                subsection_name = getattr(section_or_callable, item_key_attr)
            else:
                # If in the main dict, or is a collection, the name attribute or function/class name holds the name
                subsection_name = _get_public_name_or_special_name(section_or_object_view, accept_output_names=False)

            subsection_path_str = SectionPath.join(section_path_str, subsection_name).path_str

            if isinstance(section_or_callable, Collection):
                # not expected to hit these, for mypy
                assert isinstance(results, dict)
                assert section_or_callable.name is not None
                # got another section within this section. recursively call get section
                results[section_or_callable.name] = self._get_section(subsection_path_str)
            elif isinstance(results, dict):
                # got a non-collection, but results were defined as a dict. Should only have collections in dict
                raise ValueError(
                    f'section {section_or_object_view.name} has both collections and items, must only '
                    f'have collections if there is at least one collection. '
                    f'Got {section_or_callable} as non-collection.'
                )
            elif self._is_specific_class(section_or_callable):
                results.append(self._get_one_obj_with_config(subsection_path_str))
            elif callable(section_or_callable):
                # get function
                results.append(self._get_one_func_with_config(subsection_path_str))
            else:
                raise ValueError(f'could not get section {subsection_path_str}. expected Collection or '
                                 f'function or specific class,'
                                 f'got {section_or_callable} of type {type(section_or_callable)}')

        return results

    def _get_one_func_with_config(self, section_path_str: str) -> Callable:
        func, config_dict = self._get_func_and_config(section_path_str)

        full_func = partial(func, **config_dict)

        self._add_full_section_path_str_to_obj(section_path_str, full_func)
        self._add_to_config_dependencies_if_necessary(section_path_str)
        return full_func

    def _get_one_pipeline_with_config(self, section_path_str: str) -> Pipeline:
        pipeline_class, config_dict = self._get_pipeline_and_config(section_path_str)

        # Construct new pipeline instance with config args
        obj = pipeline_class(**config_dict)  # type: ignore

        self._add_full_section_path_str_to_obj(section_path_str, obj)
        self._add_to_config_dependencies_if_necessary(section_path_str)

        return obj

    def _get_one_obj_with_config(self, section_path_str: str) -> Type:
        if section_path_str in self._loaded_objects:
            return self._loaded_objects[section_path_str]

        klass, config_dict = self._get_class_and_config(section_path_str)

        obj = klass(**config_dict)

        self._add_full_section_path_str_to_obj(section_path_str, obj)
        self._loaded_objects[section_path_str] = obj
        self._add_to_config_dependencies_if_necessary(section_path_str)

        return obj

    def _get_config(self, section_path_str: str) -> ActiveFunctionConfig:
        config = self._config.get(section_path_str)
        if config is None:
            raise ConfigManagerNotLoadedException('no config to get')
        return config

    def _get_func_or_collection(self, section_path_str: str) -> PipelineOrFunctionOrCollection:
        section_path = SectionPath(section_path_str)
        registrar_name = section_path[0]

        # Check for specific class dict matching name
        for registrar in self._registrars:
            if registrar.name == registrar_name:
                lookup_in_registrar_section_path = SectionPath.from_section_str_list(section_path[1:]).path_str
                if not lookup_in_registrar_section_path:
                    # Was looking up registrar collection itself
                    return registrar.collection
                # Looking up within registrar
                return registrar.get(lookup_in_registrar_section_path)

        # Try to return from general registrar
        return self._general_registrar.get(section_path_str)

        # raise NoRegistrarWithNameException(registrar_name)

    def _get_func_and_config(self, section_path_str: str) -> Tuple[Callable, dict]:
        config = self._get_config(section_path_str)
        func = self._get_func_or_collection(section_path_str)
        func = cast(Callable, func)

        # Only pass items in config which are arguments of function
        config_dict = config.for_function(func)

        return func, config_dict

    def _get_pipeline_and_config(self, section_path_str: str) -> Tuple[Pipeline, dict]:
        config = self._get_config(section_path_str)
        # pipeline is an instance of the pipeline, without config
        pipeline = self._get_func_or_collection(section_path_str)
        pipeline = cast(Pipeline, pipeline)

        # TODO [#28]: verify whether pipeline is class or object in _get_pipeline_and_config
        #
        # Originally the return type from `_get_pipeline_and_config` was `Tuple[type, dict]`.
        # That didn't seem like what it was doing so updated type to Pipeline. But later
        # usage shows trying to construct an instance from a type. Type ignores have been
        # added for those lines in `_run_one_pipeline` and `_get_one_pipeline_with_config`.
        # Verify what is going on here and adjust.

        # Only pass items in config which are arguments of function
        config_dict = config.for_function(pipeline.__init__)  # type: ignore

        return pipeline, config_dict

    def _get_class_and_config(self, section_path_str: str) -> Tuple[Type, dict]:
        config = self._get_config(section_path_str)
        obj = self._get_func_or_collection(section_path_str)
        klass = obj.__class__

        # Only pass items in config which are arguments of function
        config_dict = config.for_function(klass)

        return klass, config_dict

    def update(self, d: dict=None, section_path_str: str=None, **kwargs):
        self._config.update(d, section_path_str, **kwargs)

        if section_path_str in self._loaded_objects:
            if d is None:
                d = {}
            d.update(**kwargs)
            for key, value in d.items():
                setattr(self._loaded_objects[section_path_str], key, value)

    def refresh(self, section_path_str: str):
        if section_path_str in self._loaded_objects:
            del self._loaded_objects[section_path_str]

        self._config.refresh(section_path_str)

    def _is_specific_class(self, obj: Any) -> bool:
        return self._all_specific_classes and isinstance(obj, self._all_specific_classes)  # type: ignore

    def _add_to_config_dependencies_if_necessary(self, section_path_str: str):
        from pyfileconf.context import context
        full_sp = SectionPath.join(self._manager_name, section_path_str)
        context.add_config_dependency_for_currently_running_item_if_exists(full_sp, force_update=True)

    def _add_full_section_path_str_to_obj(self, section_path_str: str, obj: Any):
        full_sp_str = SectionPath.join(self._manager_name, section_path_str).path_str
        obj._section_path_str = full_sp_str




from typing import Callable, List, Any
import warnings

from pyfileconf.exceptions.pipelinemanager import NoPipelineManagerForFilepathException
from pyfileconf.logic.get import _get_from_nested_obj_by_section_path
from pyfileconf.logic.inspect import get_caller_filepath
from pyfileconf.sectionpath.sectionpath import SectionPath
from pyfileconf.selector.logic.get.frommanager import get_pipeline_dict_path_and_specific_class_config_dicts_from_manager
from pyfileconf.pipelines.models.collection import PipelineCollection
from pyfileconf.basemodels.pipeline import Pipeline
from pyfileconf.data.models.collection import SpecificClassCollection
from pyfileconf.views.object import ObjectView

class Selector:

    def __init__(self):
        self._attach_to_pipeline_manager()
        self._load_structure()
        self._is_selector = True

    def __contains__(self, item):
        from pyfileconf.context import context

        if not isinstance(item, (str, SectionPath)):
            warnings.warn('check for if non-str object in Selector will always return False')
            return False

        collection_obj, relative_section_path = self._get_collection_obj_and_relative_section_path_from_structure(item)

        if relative_section_path is None:
            # got only the root data path, e.g. project.sources. Return the collection object itself
            return collection_obj

        # Only should return True if we find an ItemView
        # Three possible cases here.
        # Accessing a non-existent attr, should catch AttributeError then return False
        # Accessing an existing item, should be a collection, source, or function instance, return True
        # Accessing an existing attribute of an existing item, should not be an ItemView instance, return False
        try:
            result = _get_from_nested_obj_by_section_path(
                collection_obj,
                relative_section_path,
                prevent_property_access=True
            )
        except AttributeError:
            return False

        # TODO [#19]: convert all functions to pipelines
        #
        # it would make this check safer

        item_types = (Pipeline, Callable, ObjectView)
        if collection_obj.klass is not None:
            item_types = item_types + (collection_obj.klass,)
        collection_types = (SpecificClassCollection, PipelineCollection)
        pfc_types = collection_types + item_types
        if isinstance(result, pfc_types):
            if context.file_is_currently_being_loaded and isinstance(result, item_types):
                # This item is being accessed within another config
                full_sp = self._get_full_section_path_of_caller()
                # Add the config where this item is
                # being accessed as a dependent config of this item
                context.add_config_dependency(full_sp, item)
            return True

        return False


    def __getattr__(self, item):
        from pyfileconf.selector.models.itemview import ItemView
        return ItemView(item, self)

    def __call__(self, item):
        from pyfileconf.selector.models.itemview import ItemView
        return ItemView(item, self)

    def __dir__(self):
        exposed_methods = [
            'get_type'
        ]

        managers = list(self._managers.keys())

        return exposed_methods + managers

    def _get_full_section_path_of_caller(self, caller_levels: int = 4) -> SectionPath:
        # TODO [#88]: don't inspect stack to get calling section path, set in context when loading file
        from pyfileconf.main import PipelineManager

        filepath = get_caller_filepath(caller_levels=caller_levels)
        dependent_manager = PipelineManager.get_manager_by_filepath(filepath)
        dependent_sp = SectionPath.from_filepath(dependent_manager.default_config_path, filepath)
        full_sp = SectionPath.join(dependent_manager.name, dependent_sp)
        return full_sp

    def _get_dir_for_section_path(self, section_path_str: str) -> List[str]:
        collection_obj, relative_section_path = self._get_collection_obj_and_relative_section_path_from_structure(
            section_path_str
        )
        if relative_section_path is None:
            # got only the root data path, e.g. project.sources. Return the collection object itself
            return dir(collection_obj)

        result_obj = _get_from_nested_obj_by_section_path(collection_obj, relative_section_path)

        return dir(result_obj)

    def _get_collection_obj_and_relative_section_path_from_structure(self, section_path_str: str):
        # Handle accessing correct collection object.
        section_path = SectionPath(section_path_str)
        manager_name = section_path[0]

        if len(section_path) == 1:
            # Got only the manager, e.g. project
            # return only the manager itself
            return self._managers[manager_name], None

        if section_path[1] in self._managers[manager_name].specific_class_names:
            # got a specific class path
            if len(section_path) == 2:
                # got only the root data path, e.g. project.sources
                # return the collection object itself
                return self._structure[manager_name][section_path[1]], None
            collection_name = section_path[1]
            section_path_begin_index = 2
        else:
            collection_name = '_general'
            section_path_begin_index = 1

        relative_section_path = SectionPath('.'.join(section_path[section_path_begin_index:]))
        collection_obj = self._structure[manager_name][collection_name]

        return collection_obj, relative_section_path

    def get_type(self, section_path_str: str):
        collection_obj, relative_section_path = self._get_collection_obj_and_relative_section_path_from_structure(
            section_path_str
        )

        if relative_section_path is None:
            # got only the root data path, e.g. project.sources. Return the collection object itself
            return collection_obj

        # TODO [#20]: nicer error than KeyError for typo
        result = _get_from_nested_obj_by_section_path(collection_obj, relative_section_path)

        return type(result)

    def _get_real_item(self, item):
        from pyfileconf import context
        from pyfileconf.main import PipelineManager
        manager = PipelineManager.get_manager_by_section_path_str(item)
        relative_section_path = SectionPath('.'.join(SectionPath(item)[1:]))

        if context.file_is_currently_being_loaded:
            context.add_config_dependency(
                self._get_full_section_path_of_caller(caller_levels=5), item, force_update=True
            )

        return _get_from_nested_obj_by_section_path(manager, relative_section_path)

    def _set_attr_for_item(self, item: str, attr: str, value: Any):
        section_path = SectionPath(item)
        manager_name = section_path[0]
        manager = self._managers[manager_name]
        relative_section_path_str = '.'.join(section_path[1:])
        manager.config.update(
            {attr: value},
            section_path_str=relative_section_path_str
        )

    def _attach_to_pipeline_manager(self):
        from pyfileconf import context
        self._managers = context.active_managers

    def _load_structure(self):
        from pyfileconf.main import create_collections
        from pyfileconf import PipelineManager
        out_dict = {}
        manager: PipelineManager
        for manager_name, manager in self._managers.items():
            pipeline_dict_path, specific_class_name_config_dict = get_pipeline_dict_path_and_specific_class_config_dicts_from_manager(manager)
            collections, general_collection = create_collections(
                specific_class_name_config_dict,
                manager.default_config_path,
                manager.folder,
                pipeline_dict_path,
            )
            manager_dict = {
                '_general': general_collection,
            }
            for collection in collections:
                manager_dict.update({
                    collection.name: collection
                })
            out_dict[manager_name] = manager_dict
        self._structure = out_dict




def _is_selector(obj) -> bool:
    is_selector = getattr(obj, '_is_selector', False)
    return is_selector


from typing import List, Optional, Sequence, Type
import os

from pyfileconf.basemodels.container import Container
from mixins.repr import ReprMixin
from pyfileconf.imports.models.statements.container import ImportStatementContainer

StrList = List[str]

scaffolding_error = NotImplementedError('must use SpecificClassCollection or PipelineCollection, not base class Collection')

class Collection(Container, ReprMixin):

    #### Scaffolding functions. These should be overridden by collection subclasses ###

    def _set_name_map(self):
        raise scaffolding_error

    def _output_config_files(self):
        raise scaffolding_error

    def _transform_item(self, item):
        """
        Is called on each item when adding items to collection. Should handle whether the item
        is an actual item or another collection. Must return the item or collection.

        If not overridden, will just store items as is.

        Returns: item or Collection

        """
        return item

    ### base functions. These probably do not need to be overridden by collection subclasses ###

    repr_cols = ['name', 'basepath', 'items']

    def __init__(self, basepath: str, items, name: str = None,
                 imports: ImportStatementContainer = None, always_import_strs: Optional[Sequence[str]] = None,
                 always_assign_strs: Optional[Sequence[str]] = None, klass: Optional[Type] = None,
                 key_attr: str = 'name', execute_attr: str = '__call__'):
        self.basepath = basepath
        self.imports = imports
        self.name = name

        self.always_import_strs = always_import_strs
        self.always_assign_strs = always_assign_strs
        self.key_attr = key_attr
        self.execute_attr = execute_attr
        self.klass = klass
        self.items = self._transform_items(items)
        self._set_name_map()

    def __getattr__(self, item):
        try:
            return self.name_dict[item]
        except KeyError:
            raise AttributeError(f'{item}')

    def __dir__(self):
        return self.name_dict.keys()

    @property  # type: ignore
    def items(self):
        return self._items

    @items.setter
    def items(self, items):
        self._items = items
        self._set_name_map()  # need to recreate pipeline map when items change

    def append(self, item):
        self._items.append(item)
        self._set_name_map()

    def extend(self, items):
        self._items.extend(items)
        self._set_name_map()

    @classmethod
    def from_dict(cls, dict_: dict, basepath: str, name: str = None,
                  imports: ImportStatementContainer = None, always_import_strs: Optional[Sequence[str]] = None,
                  always_assign_strs: Optional[Sequence[str]] = None, klass: Optional[Type] = None,
                  key_attr: str = 'name', execute_attr: str = '__call__'):
        items = []
        for section_name, dict_or_list in dict_.items():
            section_basepath = os.path.join(basepath, section_name)
            if isinstance(dict_or_list, dict):
                # Got another pipeline dict. Recursively process
                items.append(
                    cls.from_dict(
                        dict_or_list, basepath=section_basepath, name=section_name, imports=imports,
                        always_assign_strs=always_assign_strs, always_import_strs=always_import_strs, klass=klass,
                        key_attr=key_attr, execute_attr=execute_attr
                    )
                )
            elif isinstance(dict_or_list, list):
                # Got a list of functions or pipelines. Create a collection directly from items
                items.append(
                    cls.from_list(
                        dict_or_list, basepath=section_basepath, name=section_name, imports=imports,
                        always_assign_strs=always_assign_strs, always_import_strs=always_import_strs, klass=klass,
                        key_attr=key_attr, execute_attr=execute_attr
                    )
                )

        return cls(
            basepath=basepath, items=items, name=name,
            always_assign_strs=always_assign_strs, always_import_strs=always_import_strs, klass=klass,
            key_attr=key_attr, execute_attr=execute_attr
        )

    @classmethod
    def from_list(cls, list_: list, basepath: str, name: str = None,
                  imports: ImportStatementContainer = None, always_import_strs: Optional[Sequence[str]] = None,
                  always_assign_strs: Optional[Sequence[str]] = None, klass: Optional[Type] = None,
                  key_attr: str = 'name', execute_attr: str = '__call__'):
        items = []
        for dict_or_item in list_:
            if isinstance(dict_or_item, dict):
                # Dict within list means that there is no name for the dict. Instead just access the keys
                # of the dict by their names.
                for section_name, dict_list_or_item in dict_or_item.items():
                    section_basepath = os.path.join(basepath, section_name)
                    if isinstance(dict_list_or_item, dict):
                        collection = cls.from_dict(
                            dict_list_or_item, basepath=section_basepath, name=section_name, imports=imports,
                            always_assign_strs=always_assign_strs, always_import_strs=always_import_strs, klass=klass,
                            key_attr=key_attr, execute_attr=execute_attr
                        )
                    elif isinstance(dict_list_or_item, list):
                        collection = cls.from_list(
                            dict_list_or_item, basepath=section_basepath, name=section_name, imports=imports,
                            always_assign_strs=always_assign_strs, always_import_strs=always_import_strs, klass=klass,
                            key_attr=key_attr, execute_attr=execute_attr
                        )
                    else:
                        collection = dict_list_or_item
                    items.append(collection)
            else:
                # pipeline or function
                items.append(dict_or_item)

        return cls(
            basepath=basepath, items=items, name=name, imports=imports,
            always_assign_strs=always_assign_strs, always_import_strs=always_import_strs, klass=klass,
            key_attr=key_attr, execute_attr=execute_attr
        )

    def to_nested_dict(self):
        return to_nested_dict(self)

    def _transform_items(self, items):
        return [self._transform_item(item) for item in items]


def to_nested_dict(collection: Collection):
    out_dict = {}
    for item in collection:
        if isinstance(item, Collection):
            nested = to_nested_dict(item)
            out_dict[item.name] = nested
        else:
            out_dict[item.name] = item
    return out_dict


from typing import List, Union, TYPE_CHECKING, Sequence, Optional, cast

if TYPE_CHECKING:
    from pyfileconf.selector.models.itemview import ItemView
import os

from mixins.repr import ReprMixin

SectionPathOrStr = Union[str, 'SectionPath']

class SectionPath(ReprMixin):
    repr_cols = ['path_str', 'sections']

    def __init__(self, section_path: str):
        self.path_str = section_path
        self.sections = _section_path_str_to_section_strs(section_path)

    def __iter__(self):
        for section in self.sections:
            yield section

    def __getitem__(self, item):
        return self.sections[item]

    def __len__(self):
        return len(self.sections)

    @classmethod
    def from_filepath(cls, basepath: str, filepath: str):
        relative_path = os.path.relpath(filepath, start=basepath)
        section_path = _relative_filepath_to_section_path(relative_path)
        return cls(section_path)

    @classmethod
    def from_section_str_list(cls, section_strs: List[str]):
        section_path = _section_strs_to_section_path_str(section_strs)
        return cls(section_path)

    @classmethod
    def from_ambiguous(cls, item: Union[str, 'ItemView', 'SectionPath'], strip_manager_from_iv: bool = False):
        from pyfileconf.selector.models.itemview import _is_item_view
        if _is_item_view(item):
            item = cast('ItemView', item)
            sp = cls(item.section_path_str)
            if strip_manager_from_iv:
                relative_section_path = cls.from_section_str_list(sp[1:])
                return relative_section_path
            else:
                return sp
        if isinstance(item, cls):
            return item
        if isinstance(item, str):
            return cls(item)
        raise ValueError(f'cannot parse {item} of type {item} to create SectionPath')

    @classmethod
    def list_from_ambiguous(
        cls,
        items: Union[str, 'ItemView', 'SectionPath', Sequence[Union[str, 'ItemView', 'SectionPath']]],
        base_section_path_str: Optional[str] = None,
        strip_manager_from_iv: bool = False
    ) -> List['SectionPath']:
        if isinstance(items, (list, tuple)):
            items_list = items
        else:
            items = cast(Union[str, 'ItemView', 'SectionPath'], items)
            items_list = [items]

        sp_results: List[SectionPath] = []
        for item in items_list:
            sp = SectionPath.from_ambiguous(item, strip_manager_from_iv=strip_manager_from_iv)
            if base_section_path_str is not None:
                sp = SectionPath.join(base_section_path_str, sp)
            sp_results.append(sp)

        return sp_results

    @classmethod
    def join(cls, section_path_1: SectionPathOrStr, section_path_2: SectionPathOrStr):
        section_path_1 = _convert_to_section_path_if_necessary(section_path_1)
        section_path_2 = _convert_to_section_path_if_necessary(section_path_2)

        output_path_sections = section_path_1.sections + section_path_2.sections
        return cls.from_section_str_list(output_path_sections)

    def to_filepath(self, basepath: str):
        relative_path = _section_path_to_relative_filepath(self)
        return os.path.join(basepath, relative_path)

    def endswith(self, sp: 'SectionPath'):
        num_to_match = len(sp)
        match_parts = self[-num_to_match:]
        for i, part in enumerate(match_parts):
            if part != sp[i]:
                return False
        return True

    def __eq__(self, other):
        try:
            return self.path_str == other.path_str
        except AttributeError:
            return False

    def __hash__(self):
        return hash(self.path_str)


def _section_path_str_to_section_strs(section_path_str: str) -> List[str]:
    return section_path_str.split('.')

def _section_strs_to_section_path_str(section_strs: List[str]) -> str:
    return '.'.join(section_strs)

def _relative_filepath_to_section_path(filepath: str) -> str:
    sections = filepath.split(os.path.sep)
    sections[-1] = _strip_py(sections[-1]) # remove .py if exists at end of filepath
    return _section_strs_to_section_path_str(sections)

def _section_path_to_relative_filepath(section_path: SectionPath) -> str:
    return os.path.sep.join(section_path.sections)

def _strip_py(filename_str: str) -> str:
    if not filename_str.endswith('.py'):
        return filename_str

    return filename_str[:-3]

def _convert_to_section_path_if_necessary(section_path: SectionPathOrStr) -> SectionPath:
    if isinstance(section_path, SectionPath):
        return section_path
    elif isinstance(section_path, str):
        return SectionPath(section_path)
    else:
        raise ValueError(f'expected SectionPath or str. got {section_path} of type {type(section_path)}')


from pyfileconf.exceptions.base import ManagerException


class PipelineManagerNotLoadedException(ManagerException):
    pass


class NoPipelineManagerForFilepathException(ManagerException):
    pass


class NoPipelineManagerForSectionPathException(ManagerException):
    pass


from typing import List
import ast
import importlib

from mixins.repr import ReprMixin

from pyfileconf.exceptions.imports import CouldNotExtractObjectImportFromAstException
from pyfileconf.imports.models.statements.importbase import ImportStatement
from pyfileconf.imports.models.statements.rename import RenameStatementCollection
from pyfileconf.imports.models.statements.comment import Comment
from pyfileconf.imports.logic.load.ext_importlib import get_filepath_from_module_str


class ObjectImportStatement(ImportStatement, ReprMixin):
    rename_attr = 'objs'
    repr_cols = ['module', 'objs', 'renames', 'comment']
    equal_attrs = ['module', 'objs', 'renames', 'comment']

    def __init__(self, objs: List[str], module: str, renames: RenameStatementCollection = None, comment: Comment=None,
                 preferred_position: str = None):

        if renames is None:
            renames = RenameStatementCollection([])

        self.objs = objs
        self.module = module
        self.renames = renames
        self.comment = comment
        self.preferred_position = preferred_position  # sets self.prefer_beginning as bool

    def __str__(self):
        objs = self._renamed
        objs_str = ', '.join(objs)
        import_str = f'from {self.module} import {objs_str}'
        if self.comment is not None:
            import_str += f' {self.comment}'

        return import_str

    @classmethod
    def from_str(cls, import_str: str, comment: Comment=None,
                 preferred_position: str = None):
        from pyfileconf.io.file.load.parsers.imp import extract_obj_import_from_ast
        ast_module = ast.parse(import_str)
        cls_obj = extract_obj_import_from_ast(ast_module)
        if cls_obj is None:
            raise CouldNotExtractObjectImportFromAstException(f'Original str which was converted to ast: {import_str}')
        cls_obj.comment = comment
        cls_obj.preferred_position = preferred_position

        return cls_obj

    @classmethod
    def from_ast_import_from(cls, ast_import: ast.ImportFrom, preferred_position: str = None):

        # Create RenameStatementCollection
        renames = RenameStatementCollection.from_ast_import(ast_import)

        # Collect object names
        objs = [alias.name for alias in ast_import.names]

        if ast_import.module is None:
            raise CouldNotExtractObjectImportFromAstException(f'original ast: {ast_import}')

        return cls(
            objs,
            ast_import.module,
            renames,
            preferred_position=preferred_position
        )

    def get_module_filepath(self, import_section_path_str: str=None) -> str:
        return get_filepath_from_module_str(self.module, import_section_path_str)

    def execute(self, import_section_path_str: str=None) -> list:
        module = importlib.import_module(self.module, import_section_path_str)
        return [getattr(module, obj_name) for obj_name in self.objs]


import os
from collections import defaultdict
from typing import Optional, Dict, Set
from unittest import TestCase

from pyfileconf import PipelineManager
from pyfileconf.main import create_project
from pyfileconf.sectionpath.sectionpath import SectionPath
from pyfileconf import context
from tests.input_files.amodule import SecondExampleClass, a_function
from tests.input_files.mypackage.cmodule import ExampleClass
from tests.utils import delete_project, nested_pipeline_dict_str_with_obj, pipeline_dict_str_with_obj, \
    nested_class_dict_str, class_dict_str

BASE_GENERATED_DIR = os.path.join('tests', 'generated_files')
INPUT_FILES_DIR = os.path.join('tests', 'input_files')
EC_CLASS_DICT = {
    'class': ExampleClass,
    'name': 'example_class'
}
EC_CLASS_DICT2 = {
    'class': ExampleClass,
    'name': 'example_class2'
}
SEC_CLASS_DICT = {
    'class': SecondExampleClass,
    'name': 'second_example_class'
}
CLASS_CONFIG_DICT_LIST = [
    EC_CLASS_DICT
]
SAME_CLASS_CONFIG_DICT_LIST = [
    EC_CLASS_DICT,
    EC_CLASS_DICT2
]
DIFFERENT_CLASS_CONFIG_DICT_LIST = [
    EC_CLASS_DICT,
    SEC_CLASS_DICT
]
FULL_CLASS_DICT_LIST = [
    EC_CLASS_DICT,
    EC_CLASS_DICT2,
    SEC_CLASS_DICT
]


class PipelineManagerTestBase(TestCase):
    defaults_folder_name = 'custom_defaults'
    pm_folder = os.path.join(BASE_GENERATED_DIR, 'first')
    second_pm_folder = os.path.join(BASE_GENERATED_DIR, 'second')
    defaults_path = os.path.join(pm_folder, defaults_folder_name)
    second_defaults_path = os.path.join(pm_folder, defaults_folder_name)
    pipeline_dict_path = os.path.join(pm_folder, 'pipeline_dict.py')
    second_pipeline_dict_path = os.path.join(second_pm_folder, 'pipeline_dict.py')
    example_class_file_names = [
        'example_class_dict.py',
        'example_class2_dict.py',
        'second_example_class_dict.py',
    ]
    standard_section_path = os.path.join(defaults_path, 'stuff')
    standard_a_function_path = os.path.join(standard_section_path, 'a_function.py')
    standard_ec_path = os.path.join(standard_section_path, 'ExampleClass.py')
    standard_specific_class_defaults_path = os.path.join(defaults_path, 'example_class')
    standard_specific_class_section_path = os.path.join(standard_specific_class_defaults_path, 'stuff')
    standard_specific_class_ec_path = os.path.join(standard_specific_class_section_path, 'data.py')

    example_class_dict_paths = []
    for name in example_class_file_names:
        example_class_dict_paths.append(os.path.join(pm_folder, name))
    second_example_class_dict_paths = []
    for name in example_class_file_names:
        second_example_class_dict_paths.append(os.path.join(second_pm_folder, name))
    logs_path = os.path.join(pm_folder, 'MyLogs')
    all_paths = (
        defaults_path,
        pm_folder,
        *example_class_dict_paths,
        logs_path
    )
    test_name = 'test_pipeline_manager'
    second_test_name = 'test_pipeline_manager2'
    expect_pm_1_a_function_depends_on_pm_1_specific_class = {
        'test_pipeline_manager.example_class.stuff.data': {SectionPath('test_pipeline_manager.stuff.a_function')}
    }
    expect_pm_1_a_function_depends_on_pm_2_a_function = {
        'test_pipeline_manager2.stuff.a_function': {SectionPath('test_pipeline_manager.stuff.a_function')}
    }
    expect_pm_1_a_function_depends_on_pm_2_specific_class = {
        'test_pipeline_manager2.example_class.stuff.data': {SectionPath('test_pipeline_manager.stuff.a_function')}
    }
    expect_pm_1_class_depends_on_pm_2_class = {
        'test_pipeline_manager2.stuff.ExampleClass': {SectionPath('test_pipeline_manager.stuff.ExampleClass')}
    }
    expect_pm_1_specific_class_depends_on_pm_2_specific_class = {
        'test_pipeline_manager2.example_class.stuff.data': {SectionPath('test_pipeline_manager.example_class.stuff.data')}
    }
    expect_pm_1_specific_class_depends_on_pm_1_specific_class_3_class_1_2_function_1_2 = {
        'test_pipeline_manager.example_class.stuff.data3': {SectionPath('test_pipeline_manager.example_class.stuff.data')},
        'test_pipeline_manager.ec.ExampleClass': {SectionPath('test_pipeline_manager.example_class.stuff.data')},
        'test_pipeline_manager.ec2.ExampleClass': {SectionPath('test_pipeline_manager.example_class.stuff.data')},
        'test_pipeline_manager.af.a_function': {SectionPath('test_pipeline_manager.example_class.stuff.data')},
        'test_pipeline_manager.af2.a_function': {SectionPath('test_pipeline_manager.example_class.stuff.data')},
    }
    expect_pm_1_specific_class_4_depends_on_pm_1_specific_class_3_class_1_2_function_1_2 = {
        'test_pipeline_manager.example_class.stuff.data3': {SectionPath('test_pipeline_manager.example_class.stuff.data4')},
        'test_pipeline_manager.ec.ExampleClass': {SectionPath('test_pipeline_manager.example_class.stuff.data4')},
        'test_pipeline_manager.ec2.ExampleClass': {SectionPath('test_pipeline_manager.example_class.stuff.data4')},
        'test_pipeline_manager.af.a_function': {SectionPath('test_pipeline_manager.example_class.stuff.data4')},
        'test_pipeline_manager.af2.a_function': {SectionPath('test_pipeline_manager.example_class.stuff.data4')},
    }

    def setup_method(self, method):
        create_project(self.pm_folder, self.logs_path, FULL_CLASS_DICT_LIST)
        create_project(self.second_pm_folder, self.logs_path, FULL_CLASS_DICT_LIST)

    def teardown_method(self, method):
        delete_project(self.pm_folder, self.logs_path, FULL_CLASS_DICT_LIST)
        delete_project(self.second_pm_folder, self.logs_path, FULL_CLASS_DICT_LIST)
        self.reset_pm_class()

    def create_pm(self, **kwargs):
        all_kwargs = dict(
            folder=self.pm_folder,
            name=self.test_name,
            log_folder=self.logs_path,
            default_config_folder_name=self.defaults_folder_name,
        )
        all_kwargs.update(**kwargs)
        pipeline_manager = PipelineManager(**all_kwargs)
        return pipeline_manager

    def reset_pm_class(self):
        context.reset()

    def write_a_function_to_pipeline_dict_file(self, nest_section: bool = False, file_path: Optional[str] = None):
        if file_path is None:
            file_path = self.pipeline_dict_path

        if nest_section:
            write_str = nested_pipeline_dict_str_with_obj(
                a_function, 'my_section', 'stuff', 'tests.input_files.amodule'
            )
        else:
            write_str = pipeline_dict_str_with_obj(a_function, 'stuff', 'tests.input_files.amodule')
        with open(file_path, 'w') as f:
            f.write(write_str)

    def write_example_class_to_pipeline_dict_file(self, file_path: Optional[str] = None):
        if file_path is None:
            file_path = self.pipeline_dict_path
        with open(file_path, 'w') as f:
            f.write(pipeline_dict_str_with_obj(ExampleClass, 'stuff', 'tests.input_files.mypackage.cmodule'))

    def write_example_class_dict_to_file(self, idx: int = 0, nest_section: bool = False,
                                         pm_index: Optional[int] = 0):
        if pm_index == 0:
            ecdp = self.example_class_dict_paths
        elif pm_index == 1:
            ecdp = self.second_example_class_dict_paths
        else:
            raise ValueError(f'must pass 0 or 1 for pm_index, got {pm_index}')

        file_path = ecdp[idx]
        if nest_section:
            write_str = nested_class_dict_str('class_dict', 'my_section', 'stuff', 'data')
        else:
            write_str = class_dict_str('class_dict', 'stuff', 'data')
        with open(file_path, 'w') as f:
            f.write(write_str)

    def write_error_to_file(self, file_path: str):
        directory = os.path.dirname(file_path)
        if not os.path.exists(directory):
            os.makedirs(directory)
        with open(file_path, 'w') as f:
            f.write('raise ValueError')

    def write_error_to_a_function_file(self):
        self.write_error_to_file(self.standard_a_function_path)

    def write_error_to_example_class_file(self):
        self.write_error_to_file(self.standard_ec_path)

    def write_error_to_specific_example_class_file(self):
        self.write_error_to_file(self.standard_specific_class_ec_path)

    def append_to_a_function_config(self, to_add: str):
        section_folder = os.path.join(self.defaults_path, 'stuff')
        config_path = os.path.join(section_folder, 'a_function.py')
        with open(config_path, 'a') as f:
            f.write(to_add)

    def append_to_example_class_config(self, to_add: str):
        section_folder = os.path.join(self.defaults_path, 'stuff')
        config_path = os.path.join(section_folder, 'ExampleClass.py')
        with open(config_path, 'a') as f:
            f.write(to_add)

    def append_to_specific_class_config(self, to_add: str):
        class_folder = os.path.join(self.defaults_path, 'example_class')
        section_folder = os.path.join(class_folder, 'stuff')
        config_path = os.path.join(section_folder, 'data.py')
        with open(config_path, 'a') as f:
            f.write(to_add)



class ImportException(Exception):
    pass


class NoImportStatementException(ImportException):
    pass


class CouldNotImportException(ImportException):
    pass


class NoImportMatchingNameException(ImportException):
    pass


class CouldNotExtractImportException(ImportException):
    pass


class CouldNotExtractModuleImportFromAstException(CouldNotExtractImportException):
    pass


class CouldNotExtractObjectImportFromAstException(CouldNotExtractImportException):
    pass


class CouldNotExtractRenameException(ImportException):
    pass


class CouldNotDetermineModuleForObjectException(ImportException):
    pass


class ExtractedIncorrectTypeOfImportException(ImportException):
    pass



from typing import TYPE_CHECKING, List
if TYPE_CHECKING:
    from pyfileconf.main import PipelineManager

from pyfileconf.exceptions.pipelinemanager import PipelineManagerNotLoadedException
from pyfileconf.sectionpath.sectionpath import SectionPath


def handle_pipeline_manager_not_loaded_or_typo(full_section_path_str: str, managers: List['PipelineManager']):
    manager_name = SectionPath(full_section_path_str)[0]
    if manager_name in managers:  # if manager is loaded
        # Even though manager is loaded, cannot find item. it is likely a typo.
        raise ItemNotFoundException(f'could not find item {full_section_path_str}')
    else:
        raise PipelineManagerNotLoadedException('create pipeline manager instance before using selectors')


def handle_known_typo_at_end_of_section_path_str(full_section_path_str: str):
    parts = full_section_path_str.split('.')
    raise ItemNotFoundException(f'could not find item {parts[-1]} of collection {".".join(parts[1:-1])} '
                                f'in pipeline manager {parts[0]}\n\ncheck spelling in {full_section_path_str}')


def handle_known_typo_after_pipeline_manager_name(full_section_path_str: str):
    parts = full_section_path_str.split('.')
    raise ItemNotFoundException(f'check spelling on first attribute after pipeline manager {parts[0]}')


class ItemNotFoundException(Exception):
    pass



"""
Default behavior to be run on hooks
"""
from typing import Any, Dict, Sequence, List, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from pyfileconf.iterate import IterativeRunner

from pyfileconf.plugin.impl import hookimpl


@hookimpl
def pyfileconf_iter_get_cases(
    config_updates: Sequence[Dict[str, Any]], runner: "IterativeRunner"
) -> List[Tuple[Dict[str, Any], ...]]:
    """
    Collect user passed casses by section path str then call
    itertools.product to produce the cases

    :param config_updates: list of kwarg dictionaries which would normally be provided to .update
    :return: config cases to be run
    """
    from pyfileconf.iterate import get_config_product

    return get_config_product(config_updates)



from typing import List, Optional
from mixins.attrequals import EqOnAttrsMixin
from pyfileconf.mixins.orderpref import OrderPreferenceMixin

class ImportStatement(EqOnAttrsMixin, OrderPreferenceMixin):
    rename_attr: Optional[str] = None

    ### Scaffolding methods

    def execute(self):
        # execute import statement
        raise NotImplementedError('use ModuleImportStatement or ObjectImportStatement, not base ImportStatement')

    #### Common class methods and properties

    @property
    def _renamed(self):
        if self.rename_attr is None:
            raise ValueError(f'must set rename_attr in class {self.__class__.__name__}')
        rename_items: List = getattr(self, self.rename_attr)

        if self.renames is None:
            return rename_items # no saved renames, just return original attrs

        renamed_items = []
        for item in rename_items:
            if item in self.renames:
                renamed_items.append(str(self.renames[item]))
            else:
                renamed_items.append(item)

        return renamed_items




from pyfileconf.basemodels.registrar import Registrar
from pyfileconf.data.models.collection import SpecificClassCollection

class SpecificRegistrar(Registrar):
    collection_class = SpecificClassCollection

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.klass is None:
            raise ValueError('must pass class for SpecificRegistrar')
